######################################################################
# Copyright (c) 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Parses an x12 path

An x12 path is comprised of a sequence of loop identifiers, a segment
identifier, and element position, and a composite position.

The last loop id might be a segment id.

/LOOP_1/LOOP_2
/LOOP_1/LOOP_2/SEG
/LOOP_1/LOOP_2/SEG02
/LOOP_1/LOOP_2/SEG[424]02-1
SEG[434]02-1
02-1
02

"""

import re

from pyx12.errors import X12PathError


class X12Path(object):
    """
    Interface to an x12 path
    """

    re_seg_id = '(?P<seg_id>[A-Za-z][A-Za-z0-9]{1,2})?'
    re_id_val = '(\[(?P<id_val>[A-Z0-9]+)\])?'
    re_ele_idx = '(?P<ele_idx>[0-9]{2})?'
    re_subele_idx = '(-(?P<subele_idx>[0-9]+))?'
    re_str = '^%s%s%s%s$' % (re_seg_id, re_id_val, re_ele_idx, re_subele_idx)
    rec_path = re.compile(re_str, re.S)

    def __init__(self, path_str):
        """
        @param path_str:
        @type path_str: string

        """
        #self.loop_list =
        self.seg_id = None
        self.id_val = None
        self.ele_idx = None
        self.subele_idx = None
        self.relative = None
        self.loop_list = []
        if path_str == '':
            self.relative = True
            return
        if path_str[0] == '/':
            self.relative = False
            self.loop_list = path_str[1:].split('/')
        else:
            self.relative = True
            self.loop_list = path_str.split('/')
        if len(self.loop_list) == 0:
            return
        if len(self.loop_list) > 0 and self.loop_list[-1] == '':
            # Ended in a /, so no segment
            del self.loop_list[-1]
            return
        if len(self.loop_list) > 0:
            seg_str = self.loop_list[-1]
            m = X12Path.rec_path.search(seg_str)
            if m is not None:
                self.seg_id = m.group('seg_id')
                self.id_val = m.group('id_val')
                if m.group('ele_idx') is not None:
                    self.ele_idx = int(m.group('ele_idx'))
                if m.group('subele_idx') is not None:
                    self.subele_idx = int(m.group('subele_idx'))
                del self.loop_list[-1]
                if self.seg_id is None and self.id_val is not None:
                    raise X12PathError('Path "%s" is invalid. Must specify a segment identifier with a qualifier' % (path_str))
                if self.seg_id is None and (self.ele_idx is not None or self.subele_idx is not None) and len(self.loop_list) > 0:
                    raise X12PathError('Path "%s" is invalid. Must specify a segment identifier' % (path_str))

    def is_match(self, path_str):
        pass

    def empty(self):
        """
        Is the path empty?
        @return: True if contains no path data
        @rtype: boolean
        """
        return self.relative is True and len(self.loop_list) == 0 and self.seg_id is None and self.ele_idx is None

    def _is_child_path(self, root_path, child_path):
        """
        Is the child path really a child of the root path?
        @type root_path: string
        @type child_path: string
        @return: True if a child
        @rtype: boolean
        """
        root = root_path.split('/')
        child = child_path.split('/')
        if len(root) >= len(child):
            return False
        for i in range(len(root)):
            if root[i] != child[i]:
                return False
        return True

    def __eq__(self, other):
        if isinstance(other, X12Path):
            return self.loop_list == other.loop_list and self.seg_id == other.seg_id \
                and self.id_val == other.id_val and self.ele_idx == other.ele_idx \
                and self.subele_idx == other.subele_idx and self.relative == other.relative
        return NotImplemented

    def __ne__(self, other):
        res = type(self).__eq__(self, other)
        if res is NotImplemented:
            return res
        return not res

    def __lt__(self, other):
        return NotImplemented

    __le__ = __lt__
    __le__ = __lt__
    __gt__ = __lt__
    __ge__ = __lt__

#    def __len__(self):
#        """
#        @rtype: int
#        """
#        return 1

    def __repr__(self):
        """
        @return: Formatted path
        @rtype: string
        """
        ret = ''
        if not self.relative:
            ret += '/'
        ret += '/'.join(self.loop_list)
        if self.seg_id and ret != '':
            ret += '/'
        ret += self.format_refdes()
        return ret

    def __hash__(self):
        return self.__repr__().__hash__()

    def format(self):
        """
        @rtype: string
        """
        return self.__repr__()

    def format_refdes(self):
        ret = ''
        if self.seg_id:
            ret += self.seg_id
            if self.id_val:
                ret += '[%s]' % self.id_val
        if self.ele_idx:
            ret += '%02i' % (self.ele_idx)
            if self.subele_idx:
                ret += '-%i' % self.subele_idx
        return ret

    def is_child_path(self, child_path):
        """
        Is the child path a child of this path?
        @type child_path: string
        @return: True if a child
        @rtype: boolean
        """
        root = self.format().split('/')
        child = child_path.split('/')
        if len(root) >= len(child):
            return False
        for i in range(len(root)):
            if root[i] != child[i]:
                return False
        return True



######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Apply local overrides to the current map.
Overrides defined in a xml document.

NOT IMPLEMENTED
"""


class map_override(object):
    """
    Apply local overrides to the current map. Overrides defined in a xml document.
    """
    def __init__(self, map_root, override_file, icvn, vriic, fic):
        pass

    def _set_value(self, map_root, path, variable, value):
        pass

    def _append_value(self, map_root, path, variable, value):
        pass

    def _reset_list(self, map_root, path, variable, value):
        pass



######################################################################
# Copyright (c)
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Visitor - Visits an error_handler composite
"""


class error_visitor(object):
    """
    """

    def __init__(self, fd):
        """
        Params:     fd - target file
        """
        pass

    def visit_root_pre(self, errh):
        """
        @param errh: Error handler
        @type errh: L{error_handler.err_handler}
        """
        pass

    def visit_root_post(self, errh):
        """
        @param errh: Error handler
        @type errh: L{error_handler.err_handler}
        """
        pass

    def visit_isa_pre(self, err_isa):
        pass

    def visit_isa_post(self, err_isa):
        """
        Params:     err_isa - error_isa instance
        """
        pass

    def visit_gs_pre(self, err_gs):
        pass

    def visit_gs_post(self, err_gs):
        """
        Params:     err_gs - error_gs instance
        """
        pass

    def visit_st_pre(self, err_st):
        pass

    def visit_st_post(self, err_st):
        """
        Params:     err_st - error_st instance
        """
        pass

    def visit_seg(self, err_seg):
        """
        Params:     err_seg - error_seg instance
        """
        pass

    def visit_ele(self, err_ele):
        """
        Params:     err_ele - error_ele instance
        """
        pass



######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
External Codes interface
"""

import os.path
import logging
from pkg_resources import resource_stream
import xml.etree.cElementTree as et

# Intrapackage imports
from pyx12.errors import EngineError


class CodesError(Exception):
    """Class for code modules errors."""


class ExternalCodes(object):
    """
    Validates an ID against an external list of codes
    """

    def __init__(self, base_path=None, exclude=None):
        """
        Initialize the external list of codes
        @param base_path: Override directory containing codes.xml.  If None,
            uses package resource folder
        @type base_path: string
        @param exclude: comma separated string of external codes to ignore
        @type exclude: string

        @note: self.codes - map of a tuple of two dates and a list of codes
        {codeset_id: (eff_dte, exp_dte, [code_values])}
        """
        logger = logging.getLogger('pyx12')
        self.codes = {}
        codes_file = 'codes.xml'
        if base_path is not None:
            logger.debug("Looking for codes file '{}' in map_path '{}'".format(codes_file, base_path))
            code_fd = open(os.path.join(base_path, codes_file))
        else:
            logger.debug("Looking for codes file '{}' in pkg_resources".format(codes_file))
            code_fd = resource_stream(__name__, os.path.join('map', codes_file))

        self.exclude_list = exclude.split(',') if exclude is not None else []

        for cElem in et.parse(code_fd).iter('codeset'):
            codeset_id = cElem.findtext('id')
            name = cElem.findtext('name')
            data_ele = cElem.findtext('data_ele')
            codes = []
            for code in cElem.iterfind('version/code'):
                codes.append(code.text)
            self.codes[codeset_id] = {'name': name, 'dataele':
                                      data_ele, 'codes': codes}


    def isValid(self, key, code, check_dte=None):
        """
        Is the code in the list identified by key
        @param key: the external codeset identifier
        @type key: string
        @param code: code to be verified
        @type code: string
        @param check_dte: deprecated
        @type check_dte: string
        @return: True if code is valid, False if not
        @rtype: boolean
        """
        #if not given a key, do not flag an error
        if not key:
            raise EngineError('bad key %s' % (key))
        #check the code against the list indexed by key
        else:
            if key in self.exclude_list:
                return True
            if key not in self.codes:
                raise EngineError('External Code "%s" is not defined' % (key))
            if code in self.codes[key]['codes']:
                return True
        return False


    def debug_print(self, count=10):
        """
        Debug print first <count> codes
        """
        for key in list(self.codes.keys()):
            print((self.codes[key][:count]))



######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Generates a 997 Response
Visitor - Visits an error_handler composite
"""

#from types import *
import time
import logging

# Intrapackage imports
from errors import EngineError
import error_visitor
import pyx12.segment

logger = logging.getLogger('pyx12.error_997')
logger.setLevel(logging.DEBUG)


class error_997_visitor(error_visitor.error_visitor):
    """
    Visit an error_handler composite.  Generate a 997.
    """
    def __init__(self, fd, term=('~', '*', '~', '\n')):
        """
        @param fd: target file
        @type fd: file descriptor
        @param term: tuple of x12 terminators used
        @type term: tuple(string, string, string, string)
        """
        self.fd = fd
        self.seg_term = '~'
        self.ele_term = '*'
        self.subele_term = ':'
        #self.seg_term = term[0]
        #self.ele_term = term[1]
        #self.subele_term = term[2]
        #self.eol = term[3]
        self.eol = '\n'
        self.seg_count = 0
        self.isa_control_num = None
        self.isa_seg = None
        self.gs_loop_count = 0
        self.gs_id = None
        self.gs_seg = None
        self.st_control_num = 0
        self.st_loop_count = 0

    def visit_root_pre(self, errh):
        """
        @param errh: Error handler
        @type errh: L{error_handler.err_handler}
        """
        #now = time.localtime()
        seg = errh.cur_isa_node.seg_data
        #ISA*00*          *00*          *ZZ*ENCOUNTER      *ZZ*00GR           *030425*1501*U*00401*000065350*0*T*:~
        self.isa_control_num = ('%s%s' % (time.strftime('%y%m%d'),
                                          time.strftime('%H%M')))[1:]
        icvn = seg.get_value('ISA12')
        isa_seg = pyx12.segment.Segment('ISA*00*          *00*          ',
                                        self.seg_term, self.ele_term, self.subele_term)
        isa_seg.append(seg.get_value('ISA07'))
        isa_seg.append(seg.get_value('ISA08'))
        isa_seg.append(seg.get_value('ISA05'))
        isa_seg.append(seg.get_value('ISA06'))
        isa_seg.append(time.strftime('%y%m%d'))  # Date
        isa_seg.append(time.strftime('%H%M'))  # Time
        isa_seg.append(seg.get_value('ISA11'))
        isa_seg.append(icvn)
        isa_seg.append(self.isa_control_num)  # ISA Interchange Control Number
        isa_seg.append(seg.get_value('ISA14'))
        isa_seg.append(seg.get_value('ISA15'))
        isa_seg.append(self.subele_term)
        self._write(isa_seg)
        self.isa_seg = isa_seg
        self.gs_loop_count = 0

        # GS*FA*ENCOUNTER*00GR*20030425*150153*653500001*X*004010
        seg = errh.cur_gs_node.seg_data
        gs_seg = pyx12.segment.Segment('GS', '~', '*', ':')
        gs_seg.append('FA')
        gs_seg.append(seg.get_value('GS03').rstrip())
        gs_seg.append(seg.get_value('GS02').rstrip())
        gs_seg.append(time.strftime('%Y%m%d'))
        gs_seg.append(time.strftime('%H%M%S'))
        gs_seg.append(seg.get_value('GS06'))
        gs_seg.append(seg.get_value('GS07'))
        gs_seg.append(icvn)
        self._write(gs_seg)
        self.gs_seg = gs_seg
        self.gs_id = seg.get_value('GS06')
        #self.gs_997_count = 0
        self.st_loop_count = 0
        self.gs_loop_count += 1

    def __get_isa_errors(self, err_isa):
        """
        Build list of TA1 level errors
        Only the first error is used
        """
        isa_ele_err_map = {1: '010', 2: '011', 3: '012', 4: '013', 5: '005', 6: '006',
                           7: '007', 8: '008', 9: '014', 10: '015', 11: '016', 12: '017', 13: '018',
                           14: '019', 15: '020', 16: '027'
                           }
        iea_ele_err_map = {1: '021', 2: '018'}
        err_codes = [err[0] for err in err_isa.errors]
        for elem in err_isa.elements:
            for (err_cde, err_str, bad_value) in elem.errors:
                # Ugly
                if 'ISA' in err_str:
                    err_codes.append(isa_ele_err_map[elem.ele_pos])
                elif 'IEA' in err_str:
                    err_codes.append(iea_ele_err_map[elem.ele_pos])
        # return unique codes
        return list(set(err_codes))

    def visit_root_post(self, errh):
        """
        @param errh: Error handler
        @type errh: L{error_handler.err_handler}
        """
        self._write(pyx12.segment.Segment('GE*%i*%s' % (self.st_loop_count,
                                                        self.gs_seg.get_value('GS06')), '~', '*', ':'))
        self.gs_loop_count = 1

        #pdb.set_trace()
        #TA1 segment
        err_isa = errh.cur_isa_node
        if err_isa.ta1_req == '1':
            #seg = ['TA1', err_isa.isa_trn_set_id, err_isa.orig_date, \
            #    err_isa.orig_time]
            ta1_seg = pyx12.segment.Segment('TA1', '~', '*', ':')
            ta1_seg.append(err_isa.isa_trn_set_id)
            ta1_seg.append(err_isa.orig_date)
            ta1_seg.append(err_isa.orig_time)
            err_codes = self.__get_isa_errors(err_isa)
            if err_codes:
                err_cde = err_codes[0]
                ta1_seg.append('R')
                ta1_seg.append(err_cde)
            else:
                ta1_seg.append('A')
                ta1_seg.append('000')
            self._write(ta1_seg)

        self._write(pyx12.segment.Segment('IEA*%i*%s' %
                                          (self.gs_loop_count, self.isa_control_num), '~', '*', ':'))

    def visit_isa_pre(self, err_isa):
        """
        @param err_isa: ISA Loop error handler
        @type err_isa: L{error_handler.err_isa}
        """

    def visit_isa_post(self, err_isa):
        """
        @param err_isa: ISA Loop error handler
        @type err_isa: L{error_handler.err_isa}
        """

    def visit_gs_pre(self, err_gs):
        """
        @param err_gs: GS Loop error handler
        @type err_gs: L{error_handler.err_gs}
        """
        #ST
        self.st_control_num += 1
        #seg = ['ST', '997', '%04i' % self.st_control_num]
        #self._write(seg)
        self._write(pyx12.segment.Segment('ST*997*%04i' %
                                          (self.st_control_num), '~', '*', ':'))
        self.seg_count = 0
        self.seg_count = 1
        self.st_loop_count += 1

        #AK1
        #seg = ['AK1', err_gs.fic, err_gs.gs_control_num]
        #self._write(seg)
        self._write(pyx12.segment.Segment('AK1*%s*%s' %
                                          (err_gs.fic, err_gs.gs_control_num), '~', '*', ':'))

    def __get_gs_errors(self, err_gs):
        """
        Build list of GS level errors
        """
        gs_ele_err_map = {6: '6', 8: '2'}
        ge_ele_err_map = {2: '6'}
        err_codes = [err[0] for err in err_gs.errors]
        for elem in err_gs.elements:
            for (err_cde, err_str, bad_value) in elem.errors:
                # Ugly
                if 'GS' in err_str:
                    #if elem.ele_pos in gs_ele_err_map.keys():
                    if elem.ele_pos in gs_ele_err_map:
                        err_codes.append(gs_ele_err_map[elem.ele_pos])
                    else:
                        err_codes.append('1')
                elif 'GE' in err_str:
                    if elem.ele_pos in ge_ele_err_map:
                        err_codes.append(ge_ele_err_map[elem.ele_pos])
                    else:
                        err_codes.append('1')
        # return unique codes
        ret = list(set(err_codes))
        ret.sort()
        return ret

    def visit_gs_post(self, err_gs):
        """
        @param err_gs: GS Loop error handler
        @type err_gs: L{error_handler.err_gs}
        """
        if not (err_gs.ack_code and err_gs.st_count_orig and
                err_gs.st_count_recv):
            #raise EngineError, 'err_gs variables not set'
            if not err_gs.ack_code:
                err_gs.ack_code = 'R'
            if not err_gs.st_count_orig:
                err_gs.st_count_orig = 0
            if not err_gs.st_count_recv:
                err_gs.st_count_recv = 0
        #.st_count_orig = None # AK902
        #.st_count_recv = None # AK903
        #.st_count_accept = None # AK904
        #.err_cde = [] # AK905-9

        #seg.append(err_st.ack_code)
#        if '1' in err_gs.errors: ack_code = 'R'
#        elif '2' in err_gs.errors: ack_code = 'R'
#        elif '3' in err_gs.errors: ack_code = 'R'
#        elif '4' in err_gs.errors: ack_code = 'R'
#        elif '5' in err_gs.errors: ack_code = 'E'
#        elif '6' in err_gs.errors: ack_code = 'E'
#        else: ack_code = 'A'

        seg_data = pyx12.segment.Segment('AK9', '~', '*', ':')
        seg_data.append(err_gs.ack_code)
        seg_data.append('%i' % err_gs.st_count_orig)
        seg_data.append('%i' % err_gs.st_count_recv)
        count_ok = max(err_gs.st_count_recv - err_gs.count_failed_st(), 0)
        seg_data.append('%i' % (count_ok))
        #seg = ['AK9', err_gs.ack_code, '%i' % err_gs.st_count_orig, \
        #    '%i' % err_gs.st_count_recv, \
        #    '%i' % (err_gs.st_count_recv - err_gs.count_failed_st())]
        err_codes = self.__get_gs_errors(err_gs)
        for err_cde in err_codes:
            seg_data.append('%s' % err_cde)
            #seg.append('%s' % err_cde)
        self._write(seg_data)
        #for child in err_gs.children:
            #print child.cur_line, child.seg
        #logger.info('err_gs has %i children' % len(self.children))

        #SE
        seg_count = self.seg_count + 1
        seg_data = pyx12.segment.Segment('SE', '~', '*', ':')
        seg_data.append('%i' % seg_count)
        seg_data.append('%04i' % self.st_control_num)
        #seg = ['SE', '%i' % seg_count, '%04i' % self.st_control_num]
        self._write(seg_data)

    def visit_st_pre(self, err_st):
        """
        @param err_st: ST Loop error handler
        @type err_st: L{error_handler.err_st}
        """
        seg_data = pyx12.segment.Segment('AK2', '~', '*', ':')
        seg_data.append(err_st.trn_set_id)
        seg_data.append(err_st.trn_set_control_num.strip())
        self._write(seg_data)

    def __get_st_errors(self, err_st):
        """
        Build list of ST level errors
        """
        st_ele_err_map = {1: '6', 2: '7'}
        se_ele_err_map = {1: '6', 2: '7'}
        err_codes = [err[0] for err in err_st.errors]
        if err_st.child_err_count() > 0:
            err_codes.append('5')
        for elem in err_st.elements:
            for (err_cde, err_str, bad_value) in elem.errors:
                # Ugly
                if 'ST' in err_str:
                    err_codes.append(st_ele_err_map[elem.ele_pos])
                elif 'SE' in err_str:
                    err_codes.append(se_ele_err_map[elem.ele_pos])
        # return unique codes
        ret = list(set(err_codes))
        ret.sort()
        return ret

    def visit_st_post(self, err_st):
        """
        @param err_st: ST Loop error handler
        @type err_st: L{error_handler.err_st}
        """
        if err_st.ack_code is None:
            raise EngineError('err_st.ack_cde variable not set')
#        self.ack_code = None # AK501
        seg_data = pyx12.segment.Segment('AK5', '~', '*', ':')
#        self.err_cde = [] # AK502-6
        #seg.append(err_st.ack_code)
        seg_data.append(err_st.ack_code)
        err_codes = self.__get_st_errors(err_st)
        for i in range(min(len(err_codes), 5)):
            seg_data.append(err_codes[i])
            #seg.append(err_codes[i])
        self._write(seg_data)

    def visit_seg(self, err_seg):
        """
        @param err_seg: Segment error handler
        @type err_seg: L{error_handler.err_seg}
        """
        #logger.debug('visit_deg: AK3 - ')
        #seg_base = ['AK3', err_seg.seg_id, '%i' % err_seg.seg_count]
        valid_AK3_codes = ('1', '2', '3', '4', '5', '6', '7', '8')
        seg_base = pyx12.segment.Segment('AK3', '~', '*', ':')
        seg_base.append(err_seg.seg_id)
        seg_base.append('%i' % err_seg.seg_count)
        if err_seg.ls_id:
            seg_base.append(err_seg.ls_id)
        else:
            seg_base.append('')
        seg_str = seg_base.format('~', '*', ':')
        errors = [x[0] for x in err_seg.errors]
        if 'SEG1' in errors:
            if '8' not in errors:
                errors.append('8')
            errors = [x for x in errors if x != 'SEG1']
        for err_cde in list(set(errors)):
            if err_cde in valid_AK3_codes:  # unique codes
                seg_data = pyx12.segment.Segment(seg_str, '~', '*', ':')
                seg_data.set('AK304', err_cde)
                self._write(seg_data)
        if err_seg.child_err_count() > 0 and '8' not in errors:
            seg_data = pyx12.segment.Segment(seg_str, '~', '*', ':')
            seg_data.set('AK304', '8')
            self._write(seg_data)

    def visit_ele(self, err_ele):
        """
        @param err_ele: Segment error handler
        @type err_ele: L{error_handler.err_ele}
        """
        valid_AK4_codes = ('1', '2', '3', '4', '5', '6', '7', '8', '9', '10')
        seg_base = pyx12.segment.Segment('AK4', '~', '*', ':')
        if err_ele.subele_pos:
            seg_base.append('%i:%i' % (err_ele.ele_pos, err_ele.subele_pos))
        else:
            seg_base.append('%i' % (err_ele.ele_pos))
        if err_ele.ele_ref_num:
            seg_base.append(err_ele.ele_ref_num)
        #else:
        #    seg_base.append('')
        seg_str = seg_base.format('~', '*', ':')
        for (err_cde, err_str, bad_value) in err_ele.errors:
            if err_cde in valid_AK4_codes:
                seg_data = pyx12.segment.Segment(seg_str, '~', '*', ':')
                seg_data.set('AK403', err_cde)
                if bad_value:
                    seg_data.set('AK404', bad_value)
                self._write(seg_data)

    def _write(self, seg_data):
        """
        Params:     seg_data -
        @param seg_data: Data segment instance
        @type seg_data: L{segment.Segment}
        """
        sout = seg_data.format(self.seg_term, self.ele_term, self.subele_term)
        if seg_data.get_seg_id() == 'ISA':
            sout = sout[:-1] + self.ele_term + self.subele_term \
                + self.seg_term
        self.fd.write('%s\n' % (sout))
        self.seg_count += 1



######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Generates a 999 Response
Visitor - Visits an error_handler composite
"""

import time
import logging
import random

# Intrapackage imports
from pyx12.errors import EngineError
import pyx12.error_visitor
import pyx12.segment
import pyx12.x12file

logger = logging.getLogger('pyx12.error_999')
logger.setLevel(logging.DEBUG)


class error_999_visitor(pyx12.error_visitor.error_visitor):
    """
    Visit an error_handler composite.  Generate a 999.
    """
    def __init__(self, fd, term=('~', '*', ':', '\n', '^')):
        """
        @param fd: target file
        @type fd: file descriptor
        @param term: tuple of x12 terminators used
        @type term: tuple(string, string, string, string)
        """
        self.fd = fd
        self.wr = pyx12.x12file.X12Writer(fd, '~', '*', ':', '\n', '^')
        self.seg_term = '~'
        self.ele_term = '*'
        self.subele_term = ':'
        self.repetition_term = '^'
        self.eol = '\n'
        self.isa_control_num = None
        self.gs_control_num = None
        self.st_control_num = 0
        self.vriic = '005010X231'


    def visit_root_pre(self, errh):
        """
        @param errh: Error handler
        @type errh: L{error_handler.err_handler}

        Uses:
        isa_node seg_data
        gs_node seg_data
        """
        seg = errh.cur_isa_node.seg_data
        #ISA*00*          *00*          *ZZ*ENCOUNTER      *ZZ*00GR           *030425*1501*U*00501*000065350*0*T*:~
        self.isa_control_num = ('%s%s' % (time.strftime('%y%m%d'),
                                          time.strftime('%H%M')))[1:]
        self.gs_control_num = '%i' % (random.randint(10000000, 999999999))
        icvn = seg.get_value('ISA12')
        isa_seg = pyx12.segment.Segment('ISA*00*          *00*          ',
                                        self.seg_term, self.ele_term, self.subele_term)
        isa_seg.set('05', seg.get_value('ISA07'))
        isa_seg.set('06', seg.get_value('ISA08'))
        isa_seg.set('07', seg.get_value('ISA05'))
        isa_seg.set('08', seg.get_value('ISA06'))
        isa_seg.set('09', time.strftime('%y%m%d'))  # Date
        isa_seg.set('10', time.strftime('%H%M'))  # Time
        isa_seg.set('11', self.repetition_term)
        isa_seg.set('12', icvn)
        isa_seg.set('13', self.isa_control_num)  # ISA Interchange Control Number
        isa_seg.set('14', seg.get_value('ISA14'))
        isa_seg.set('15', seg.get_value('ISA15'))
        isa_seg.set('16', self.subele_term)
        self.wr.Write(isa_seg)

        # GS*FA*ENCOUNTER*00GR*20030425*150153*653500001*X*005010
        seg = errh.cur_gs_node.seg_data
        gs_seg = pyx12.segment.Segment('GS', '~', '*', ':')
        gs_seg.set('01', 'FA')
        gs_seg.set('02', seg.get_value('GS03').rstrip())
        gs_seg.set('03', seg.get_value('GS02').rstrip())
        gs_seg.set('04', time.strftime('%Y%m%d'))
        gs_seg.set('05', time.strftime('%H%M%S'))
        gs_seg.set('06', self.gs_control_num)
        gs_seg.set('07', seg.get_value('GS07'))
        gs_seg.set('08', self.vriic)
        self.wr.Write(gs_seg)

    def __get_isa_errors(self, err_isa):
        """
        Build list of TA1 level errors
        Only the first error is used
        """
        isa_ele_err_map = {1: '010', 2: '011', 3: '012', 4: '013', 5: '005', 6: '006',
                           7: '007', 8: '008', 9: '014', 10: '015', 11: '016', 12: '017', 13: '018',
                           14: '019', 15: '020', 16: '027'
                           }
        iea_ele_err_map = {1: '021', 2: '018'}
        err_codes = [err[0] for err in err_isa.errors]
        for elem in err_isa.elements:
            for (err_cde, err_str, bad_value) in elem.errors:
                # Ugly
                if 'ISA' in err_str:
                    err_codes.append(isa_ele_err_map[elem.ele_pos])
                elif 'IEA' in err_str:
                    err_codes.append(iea_ele_err_map[elem.ele_pos])
        # return unique codes
        return list(set(err_codes))

    def visit_root_post(self, errh):
        """
        @param errh: Error handler
        @type errh: L{error_handler.err_handler}
        """
        ge = pyx12.segment.Segment('GE', '~', '*', ':')
        ge.set('02', self.gs_control_num)
        self.wr.Write(ge)

        #TA1 segment
        err_isa = errh.cur_isa_node
        if err_isa.ta1_req == '1':
            #seg = ['TA1', err_isa.isa_trn_set_id, err_isa.orig_date, \
            #    err_isa.orig_time]
            ta1_seg = pyx12.segment.Segment('TA1', '~', '*', ':')
            ta1_seg.append(err_isa.isa_trn_set_id)
            ta1_seg.append(err_isa.orig_date)
            ta1_seg.append(err_isa.orig_time)
            err_codes = self.__get_isa_errors(err_isa)
            if err_codes:
                err_cde = err_codes[0]
                ta1_seg.append('R')
                ta1_seg.append(err_cde)
            else:
                ta1_seg.append('A')
                ta1_seg.append('000')
            self.wr.Write(ta1_seg)
        self.wr.Write(pyx12.segment.Segment('IEA', '~', '*', ':'))

    def visit_isa_pre(self, err_isa):
        """
        @param err_isa: ISA Loop error handler
        @type err_isa: L{error_handler.err_isa}
        """

    def visit_isa_post(self, err_isa):
        """
        @param err_isa: ISA Loop error handler
        @type err_isa: L{error_handler.err_isa}
        """
        pass

    def visit_gs_pre(self, err_gs):
        """
        @param err_gs: GS Loop error handler
        @type err_gs: L{error_handler.err_gs}
        """
        #ST
        self.st_control_num += 1
        st_seg = pyx12.segment.Segment('ST*999', '~', '*', ':')
        st_seg.set('02', '%04i' % (self.st_control_num))
        st_seg.set('03', self.vriic)
        self.wr.Write(st_seg)
        ak1 = pyx12.segment.Segment('AK1', '~', '*', ':')
        ak1.set('01', err_gs.fic)
        ak1.set('02', err_gs.gs_control_num)
        ak1.set('03', err_gs.vriic)
        self.wr.Write(ak1)

    def __get_gs_errors(self, err_gs):
        """
        Build list of GS level errors
        """
        gs_ele_err_map = {6: '6', 8: '2'}
        ge_ele_err_map = {2: '6'}
        err_codes = [err[0] for err in err_gs.errors]
        for elem in err_gs.elements:
            for (err_cde, err_str, bad_value) in elem.errors:
                # Ugly
                if 'GS' in err_str:
                    if elem.ele_pos in gs_ele_err_map:
                        err_codes.append(gs_ele_err_map[elem.ele_pos])
                    else:
                        err_codes.append('1')
                elif 'GE' in err_str:
                    if elem.ele_pos in ge_ele_err_map:
                        err_codes.append(ge_ele_err_map[elem.ele_pos])
                    else:
                        err_codes.append('1')
        # return unique codes
        ret = list(set(err_codes))
        ret.sort()
        return ret

    def visit_gs_post(self, err_gs):
        """
        @param err_gs: GS Loop error handler
        @type err_gs: L{error_handler.err_gs}
        """
        if not (err_gs.ack_code and err_gs.st_count_orig and
                err_gs.st_count_recv):
            if not err_gs.ack_code:
                err_gs.ack_code = 'R'
            if not err_gs.st_count_orig:
                err_gs.st_count_orig = 0
            if not err_gs.st_count_recv:
                err_gs.st_count_recv = 0

        seg_data = pyx12.segment.Segment('AK9', '~', '*', ':')
        seg_data.set('01', err_gs.ack_code)
        seg_data.set('02', '%i' % err_gs.st_count_orig)
        seg_data.set('03', '%i' % err_gs.st_count_recv)
        count_ok = max(err_gs.st_count_recv - err_gs.count_failed_st(), 0)
        seg_data.set('04', '%i' % (count_ok))
        err_codes = self.__get_gs_errors(err_gs)
        for err_cde in err_codes[:5]:
            seg_data.append(err_cde)
        self.wr.Write(seg_data)

        #SE
        seg_data = pyx12.segment.Segment('SE', '~', '*', ':')
        seg_data.append('%i' % (0))
        seg_data.append('%04i' % self.st_control_num)
        self.wr.Write(seg_data)

    def visit_st_pre(self, err_st):
        """
        @param err_st: ST Loop error handler
        @type err_st: L{error_handler.err_st}
        """
        if err_st is None:
            raise EngineError('Cannot create AK2 : err_st is None')
        if err_st.trn_set_id is None:
            raise EngineError('Cannot create AK2: err_st.trn_set_id was not set')
        if err_st.trn_set_control_num is None:
            raise EngineError('Cannot create AK2: err_st.trn_set_control_num was not set')
        if err_st.vriic is None:
            raise EngineError('Cannot create AK2: err_st.vriic was not set')
        seg_data = pyx12.segment.Segment('AK2', '~', '*', ':')
        seg_data.set('01', err_st.trn_set_id)
        seg_data.set('02', err_st.trn_set_control_num.strip())
        seg_data.set('03', err_st.vriic)
        self.wr.Write(seg_data)

    def __get_st_errors(self, err_st):
        """
        Build list of ST level errors
        """
        st_ele_err_map = {1: '6', 2: '7'}
        se_ele_err_map = {1: '6', 2: '7'}
        err_codes = [err[0] for err in err_st.errors]
        if err_st.child_err_count() > 0:
            err_codes.append('5')
        for elem in err_st.elements:
            for (err_cde, err_str, bad_value) in elem.errors:
                # Ugly
                if 'ST' in err_str:
                    err_codes.append(st_ele_err_map[elem.ele_pos])
                elif 'SE' in err_str:
                    err_codes.append(se_ele_err_map[elem.ele_pos])
        # return unique codes
        ret = list(set(err_codes))
        ret.sort()
        return ret

    def visit_st_post(self, err_st):
        """
        @param err_st: ST Loop error handler
        @type err_st: L{error_handler.err_st}
        """
        if err_st.ack_code is None:
            raise EngineError('err_st.ack_cde variable not set')
        seg_data = pyx12.segment.Segment('IK5', '~', '*', ':')
        seg_data.set('01', err_st.ack_code)
        err_codes = self.__get_st_errors(err_st)
        for err_code in err_codes[:5]:
            seg_data.append(err_code)
        self.wr.Write(seg_data)

    def visit_seg(self, err_seg):
        """
        @param err_seg: Segment error handler
        @type err_seg: L{error_handler.err_seg}
        """
        valid_IK3_codes = ('1', '2', '3', '4', '5', '6', '7', '8', 'I4', 'I6', 'I7', 'I8', 'I9')
        seg_base = pyx12.segment.Segment('IK3', '~', '*', ':')
        seg_base.set('01', err_seg.seg_id)
        seg_base.set('02', '%i' % err_seg.seg_count)
        if err_seg.ls_id:
            seg_base.set('03', err_seg.ls_id)
        #else:
        #    seg_base.set('')
        seg_str = seg_base.format('~', '*', ':')
        errors = [x[0] for x in err_seg.errors]
        if 'SEG1' in errors:
            if '8' not in errors:
                errors.append('8')
            errors = [x for x in errors if x != 'SEG1']
        for err_cde in list(set(errors)):
            if err_cde in valid_IK3_codes:  # unique codes
                seg_data = pyx12.segment.Segment(seg_str, '~', '*', ':')
                seg_data.set('IK304', err_cde)
                self.wr.Write(seg_data)
# todo: add segment context
# todo: add business unit context
        if err_seg.child_err_count() > 0 and '8' not in errors:
            seg_data = pyx12.segment.Segment(seg_str, '~', '*', ':')
            seg_data.set('IK304', '8')
            self.wr.Write(seg_data)

    def visit_ele(self, err_ele):
        """
        @param err_ele: Segment error handler
        @type err_ele: L{error_handler.err_ele}
        """
        valid_IK4_codes = ('1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '12', '13',
                           'I10', 'I11', 'I12', 'I13', 'I6', 'I9')
        seg_base = pyx12.segment.Segment('IK4', '~', '*', ':')
        seg_base.set('01-1', '%i' % (err_ele.ele_pos))
        if err_ele.subele_pos:
            seg_base.set('01-2', '%i' % (err_ele.subele_pos))
        if err_ele.repeat_pos:
            seg_base.set('01-3', '%i' % (err_ele.repeat_pos))
        if err_ele.ele_ref_num:
            seg_base.set('02', err_ele.ele_ref_num)
        seg_str = seg_base.format('~', '*', ':')
        for (err_cde, err_str, bad_value) in err_ele.errors:
            if err_cde in valid_IK4_codes:
                seg_data = pyx12.segment.Segment(seg_str, '~', '*', ':')
                seg_data.set('IK403', err_cde)
                if bad_value:
                    seg_data.set('IK404', bad_value)
# todo: add element context
                self.wr.Write(seg_data)



######################################################################
# Copyright (c) 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Generates error debug output
Visitor - Visits an error_handler composite
"""

# Intrapackage imports
from error_visitor import error_visitor


class error_debug_visitor(error_visitor):
    """

    """
    def __init__(self, fd):
        """
        @param fd: target file
        @type fd: file descriptor
        """
        self.fd = fd
        self.seg_count = 0
        self.st_control_num = 0

    def visit_root_pre(self, errh):
        """
        @param errh: Error_handler instance
        @type errh: L{error_handler.err_handler}
        """
        self.fd.write('%s\n' % errh.id)

    def visit_root_post(self, errh):
        """
        @param errh: Error_handler instance
        @type errh: L{error_handler.err_handler}
        """
        pass

    def visit_isa_pre(self, err_isa):
        """
        @param err_isa: ISA Loop error handler
        @type err_isa: L{error_handler.err_isa}
        """
        self.fd.write('%s\n' % err_isa.id)
        self.fd.write('-- ISA errors --\n')
        for err in err_isa.errors:
            self.fd.write('  %s %s\n' % err)
        for ele in err_isa.elements:
            self.fd.write('  %s %s\n' % (ele.id, ele.name))
            print((ele.parent))
            for err in ele.errors:
                self.fd.write('    ERR %s %s (%s)\n' % err)

    def visit_isa_post(self, err_isa):
        """
        @param err_isa: ISA Loop error handler
        @type err_isa: L{error_handler.err_isa}
        """
        pass

    def visit_gs_pre(self, err_gs):
        """
        @param err_gs: GS Loop error handler
        @type err_gs: L{error_handler.err_gs}
        """
        self.fd.write('%s\n' % err_gs.id)
        self.fd.write('-- GS errors --\n')
        for err in err_gs.errors:
            self.fd.write('  %s %s\n' % err)
        for ele in err_gs.elements:
            self.fd.write('  %s %s\n' % (ele.id, ele.name))
            for err in ele.errors:
                self.fd.write('    ERR %s %s (%s)\n' % err)

    def visit_gs_post(self, err_gs):
        """
        @param err_gs: GS Loop error handler
        @type err_gs: L{error_handler.err_gs}
        """

        self.fd.write('%s POST\n' % err_gs.id)
        #AK9
        #seg = ['AK9', err_gs.ack_code, '%i' % err_gs.st_count_orig, \
        #    '%i' % err_gs.st_count_recv, '%i' % (err_gs.st_count_recv - err_gs.count_failed_st())]
        self.fd.write(' GS Ack Code%s\n' % err_gs.ack_code)
        self.fd.write(' GS st_count_orig%s\n' % err_gs.st_count_orig)
        self.fd.write(' GS st_count_recv%i\n' % err_gs.st_count_recv)
        self.fd.write(' GS st_count_accept%i\n' % (
            err_gs.st_count_recv - err_gs.count_failed_st()))

    def visit_st_pre(self, err_st):
        """
        @param err_st: ST Loop error handler
        @type err_st: L{error_handler.err_st}
        """
        self.fd.write('%s\n' % err_st.id)
        self.fd.write('-- ST errors --\n')
        for err in err_st.errors:
            self.fd.write('  ERR %s %s\n' % err)
        for ele in err_st.elements:
            self.fd.write('  ST Element:  %s %s\n' % (ele.id, ele.name))
            for err in ele.errors:
                self.fd.write('    ERR %s %s (%s)\n' % err)

    def visit_st_post(self, err_st):
        """
        @param err_st: ST Loop error handler
        @type err_st: L{error_handler.err_st}
        """
        pass

    def visit_seg(self, err_seg):
        """
        @param err_seg: Segment error handler
        @type err_seg: L{error_handler.err_seg}
        """
        #pdb.set_trace()
        self.fd.write('%s %s %s %s\n' % (err_seg.id, err_seg.name,
                                         err_seg.get_cur_line(), err_seg.seg_id))
        for (err_cde, err_str, err_value) in err_seg.errors:
            self.fd.write('  ERR %s (%s) "%s" \n' % (err_cde,
                                                     err_value, err_str))
        #for ele in err_seg.elements:
        #    self.fd.write('  %s %s\n' % (ele.id, ele.name))

    def visit_ele(self, err_ele):
        """
        Params:     err_ele - error_ele instance
        @param err_ele: Element error handler
        @type err_ele: L{error_handler.err_ele}
        """
        self.fd.write('  %s %s\n' % (err_ele.id, err_ele.name))
        for err in err_ele.errors:
            self.fd.write('    ERR %s %s (%s)\n' % err)



# A simple XML-generator# Originally Lars Marius Garshol, September 1998
# http://mail.python.org/pipermail/xml-sig/1998-September/000347.html
# Changes by Uche Ogbuji April 2003
# *  unicode support: accept encoding argument and use Python codecs
#    for correct character output
# *  switch from deprecated string module to string methods
# *  use PEP 8 style

import sys
#import codecs


class XMLWriter(object):
    """
    Doctest:

        >>>from xmlwriter import XMLWriter
        >>>writer = XMLWriter()
        >>>writer.doctype(
        ... u"xsa", u"-//LM Garshol//DTD XML Software Autoupdate 1.0//EN//XML",
        ... u"http://www.garshol.priv.no/download/xsa/xsa.dtd")
        >>>#Notice: there is no error checking to ensure that the root element
        >>>#specified in the doctype matches the top-level element generated
        >>>writer.push(u"xsa")
        >>>#Another element with child elements
        >>>writer.push(u"vendor")
        >>>#Element with simple text (#PCDATA) content
        >>>writer.elem(u"name", u"Centigrade systems")
        >>>writer.elem(u"email", u"info@centigrade.bogus")
        >>>writer.elem(u"vendor", u"Centigrade systems")
        >>>#Close currently open element ("vendor)
        >>>writer.pop()
        >>>#Element with an attribute
        >>>writer.push(u"product", {u"id": u"100\u00B0"})
        >>>writer.elem(u"name", u"100\u00B0 Server")
        >>>writer.elem(u"version", u"1.0")
        >>>writer.elem(u"last-release", u"20030401")
        >>>#Empty element
        >>>writer.empty(u"changes")
        >>>writer.pop()
        >>>writer.pop()

    startElement
    endElement
    emptyElement
    text, data
    endDocument
    attribute
    indentation
    close
    flush
    """

    def __init__(self, out=sys.stdout, encoding="utf-8", indent=" "):
        """
        out      - a stream for the output
        encoding - an encoding used to wrap the output for unicode
        indent   - white space used for indentation
        """
        #wrapper = codecs.lookup(encoding).streamwriter
        #self.out = wrapper(out)
        self.encoding = encoding
        self.out = out
        self.stack = []
        self.indent = indent
        self._write('<?xml version="1.0" encoding="%s"?>\n' % encoding)

    def doctype(self, root, pubid, sysid):
        """
        Create a document type declaration (no internal subset)
        """
        if pubid is None:
            self._write(
                "<!DOCTYPE %s SYSTEM '%s'>\n" % (root, sysid))
        else:
            self._write(
                "<!DOCTYPE %s PUBLIC '%s' '%s'>\n"
                % (root, pubid, sysid))

    def push(self, elem, attrs={}):
        """
        Create an element which will have child elements
        """
        self._indent()
        self._write("<" + elem)
        for (a, v) in list(attrs.items()):
            self._write(" %s='%s'" % (a, self._escape_attr(v)))
        self._write(">\n")
        self.stack.append(elem)

    def elem(self, elem, content, attrs={}):
        """
        Create an element with text content only
        """
        self._indent()
        self._write("<" + elem)
        for (a, v) in list(attrs.items()):
            self._write(" %s='%s'" % (a, self._escape_attr(v)))
        self._write(">%s</%s>\n" % (self._escape_cont(content), elem))

    def empty(self, elem, attrs={}):
        """
        Create an empty element
        """
        self._indent()
        self._write("<" + elem)
        for a in list(attrs.items()):
            self._write(" %s='%s'" % a)
        self.out.write("/>\n")

    def pop(self):
        """
        Close an element started with the push() method
        """
        if len(self.stack) > 0:
            elem = self.stack[-1]
            del self.stack[-1]
            self._indent()
            self._write("</%s>\n" % elem)

    def __len__(self):
        return len(self.stack)

    def _indent(self):
        self._write(self.indent * (len(self.stack) * 2))

    def _escape_cont(self, text):
        if text is None:
            return None
        return text.replace("&", "&amp;")\
            .replace("<", "&lt;").replace(">", "&gt;")

    def _escape_attr(self, text):
        if text is None:
            return None
        return text.replace("&", "&amp;") \
            .replace("'", "&apos;").replace("<", "&lt;")\
            .replace(">", "&gt;")

    def _write(self, strval):
        self.out.write(strval.decode(self.encoding))



def dump_args(func):
    "This decorator dumps out the arguments passed to a function before calling it"
    argnames = func.func_code.co_varnames[:func.func_code.co_argcount]
    fname = func.func_name

    def echo_func(*args, **kwargs):
        print(fname, ":", ', '.join('%s=%r' % entry
            for entry in zip(argnames, args) + kwargs.items()))
        return func(*args, **kwargs)

    return echo_func



#!/usr/bin/env python

######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Parse a ANSI X12N data file.
Validate against a map and codeset values.
Create a XML document based on the data file.
"""

import os
import os.path
from os.path import isdir, isfile
import sys
import logging
import argparse

# Intrapackage imports
libpath = os.path.abspath(os.path.join(os.path.dirname(__file__), '../..'))
if os.path.isdir(libpath):
    sys.path.insert(0, libpath)
import pyx12
import pyx12.x12n_document
import pyx12.params

#Global Variables
__author__ = pyx12.__author__
__status__ = pyx12.__status__
__version__ = pyx12.__version__
__date__ = pyx12.__date__


def check_map_path_arg(map_path):
    if not isdir(map_path):
        raise argparse.ArgumentError(None, "The MAP_PATH '{}' is not a valid directory".format(map_path))
    index_file = 'maps.xml'
    if not isfile(os.path.join(map_path, index_file)):
        raise argparse.ArgumentError(None,
                    "The MAP_PATH '{}' does not contain the map index file '{}'".format(map_path, index_file))
    return map_path


def main():
    """Script main program."""
    parser = argparse.ArgumentParser(description='X12 to XML conversion')
    parser.add_argument('--config-file', '-c', action='store',
                        dest="configfile", default=None)
    parser.add_argument('--log-file', '-l', action='store', dest="logfile", default=None)
    parser.add_argument('--map-path', '-m', action='store', dest="map_path", default=None, type=check_map_path_arg)
    parser.add_argument('--verbose', '-v', action='count')
    parser.add_argument('--debug', '-d', action='store_true')
    parser.add_argument('--quiet', '-q', action='store_true')
    parser.add_argument('--html', '-H', action='store_true')
    parser.add_argument('--outputfile', '-o', action='store', help="XML target filename")
    parser.add_argument('--exclude-external-codes', '-x', action='append', dest="exclude_external",
                        default=[], help='External Code Names to ignore')
    parser.add_argument('--charset', '-s', choices=(
        'b', 'e'), help='Specify X12 character set: b=basic, e=extended')
    #parser.add_argument('--background', '-b', action='store_true')
    #parser.add_argument('--test', '-t', action='store_true')
    parser.add_argument('--version', action='version', version='{prog} {version}'.format(prog=parser.prog, version=__version__))
    parser.add_argument('input_file')
    args = parser.parse_args()

    logger = logging.getLogger('pyx12')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')

    stdout_hdlr = logging.StreamHandler()
    stdout_hdlr.setFormatter(formatter)
    logger.addHandler(stdout_hdlr)
    logger.setLevel(logging.INFO)

    param = pyx12.params.params(args.configfile)
    if args.debug:
        logger.setLevel(logging.DEBUG)
        param.set('debug', True)
    if args.verbose > 0:
        logger.setLevel(logging.DEBUG)
    if args.quiet:
        logger.setLevel(logging.ERROR)

    param.set('exclude_external_codes', ','.join(args.exclude_external))
    if args.map_path:
        param.set('map_path', args.map_path)

    if args.logfile:
        try:
            hdlr = logging.FileHandler(args.logfile)
            hdlr.setFormatter(formatter)
            logger.addHandler(hdlr)
        except IOError:
            logger.exception('Could not open log file: %s' % (args.logfile))

    if args.input_file:
        try:
            fd_src = args.input_file
        except:
            logger.error('Could not open file %s' % (args.input_file))
            return False
    else:
        fd_src = sys.stdin
    if args.outputfile:
        try:
            fd_xml = open(args.outputfile, 'w')
        except:
            logger.error('Could not open file %s' % (args.outputfile))
            return False
    else:
        fd_xml = sys.stdout
    try:
        result = pyx12.x12n_document.x12n_document(param=param, src_file=fd_src,
            fd_997=None, fd_html=None, fd_xmldoc=fd_xml, map_path=args.map_path)
        if not result:
            logger.error('Input file had errors.')
            return False
    except KeyboardInterrupt:
        print("\n[interrupt]")

    return True

if __name__ == '__main__':
    sys.exit(not main())



######################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Generates HTML error output
"""

import time
import logging
from types import ListType

# Intrapackage imports

logger = logging.getLogger('pyx12.error_html')
logger.setLevel(logging.DEBUG)
#logger.setLevel(logging.ERROR)


class error_html(object):
    """
    """
    def __init__(self, errh, fd, term=('~', '*', '~', '\n')):
        """
        @param fd: target file
        @type fd: file descriptor
        @param term: tuple of x12 terminators used
        @type term: tuple(string, string, string, string)

        @bug: GS errors are re-printing at the GE level
        """
        self.errh = errh
        self.fd = fd
        self.seg_term = term[0]
        self.ele_term = term[1]
        self.subele_term = term[2]
        self.eol = ''
        self.last_line = 0
        self.loop_info = None

    def header(self):
        self.fd.write('<html>\n<head>\n')
        self.fd.write('<title>X12N Error Analysis</title>\n')
        self.fd.write('<style type="text/css">\n<!--\n')
        self.fd.write('  span.seg { color: black; font-style: normal; }\n')
        self.fd.write('  span.error { background-color: #CCCCFF; color: red; font-style: normal; }\n')
        self.fd.write('  span.info { color: blue; font-style: normal; }\n')
        self.fd.write('  span.ele_err { background-color: yellow; color: red; font-style: normal; }\n')
        self.fd.write('  -->\n</style>\n')
        self.fd.write('  <link rel="stylesheet" href="errors.css" type="text/css" />\n')
        self.fd.write('</head>\n<body>\n')
        self.fd.write('<h1>X12N Error Analysis</h1>\n<h3>Analysis Date: %s</h3><p>\n' %
                      (time.strftime('%m/%d/%Y %H:%M:%S')))
        self.fd.write('<div class="segs" style="">\n')

    def footer(self):
        err_st = self.errh.cur_st_node
        if not err_st.is_closed():
            for (err_cde, err_str) in err_st.errors:
                if err_cde == '2':
                    self.fd.write('<span class="error">&nbsp;%s (Segment Error Code: %s)</span><br />\n' %
                                  (err_str, err_cde))
        err_gs = self.errh.cur_gs_node
        if not err_gs.is_closed():
            for (err_cde, err_str) in err_gs.errors:
                if err_cde == '3':
                    self.fd.write('<span class="error">&nbsp;%s (Segment Error Code: %s)</span><br />\n' %
                                  (err_str, err_cde))
        err_isa = self.errh.cur_isa_node
        if not err_isa.is_closed():
            for (err_cde, err_str) in err_isa.errors:
                if err_cde == '023':
                    self.fd.write('<span class="error">&nbsp;%s (Segment Error Code: %s)</span><br />\n' %
                                  (err_str, err_cde))
        self.fd.write('</div>\n')
        self.fd.write('<p>\n<a href="http://sourceforge.net/projects/pyx12/">pyx12 Validator</a>\n</p>\n')
        self.fd.write('</body>\n</html>\n')

    def loop(self, loop_node):
        if loop_node.type != 'wrapper':
            #self.gen_info('Loop %s: %s' % (loop_node.id, loop_node.name))
            self.loop_info = 'Loop %s: %s' % (loop_node.id, loop_node.name)

    def gen_info(self, info_str):
        """
        """
        self.fd.write('<span class="info">&nbsp;&nbsp;%s</span><br />\n' %
                      (info_str))

    def gen_seg(self, seg_data, src, err_node_list):
        """
        Find error seg for this segment.
        Find any skipped error values.
        ID pos of bad value.
        @param seg_data: data segment instance
        """
        cur_line = src.cur_line

        #while errh
        ele_pos_map = {}
        for err_node in err_node_list:
            for ele in err_node.elements:
                ele_pos_map[ele.ele_pos] = ele.subele_pos

        t_seg = []  # list of formatted elements
        #seg_data.format_ele_list(t_seg)
        for i in range(1, len(seg_data) + 1):
            if seg_data.is_composite(ref_des='%02i' % (i)):
                #if seg_data.get_seg_id()=='CLM': pdb.set_trace()
                t_seg.append([])
                for j in range(1, seg_data.ele_len('%02i' % (i)) + 1):
                    ref_des = '%02i-%i' % (i, j)
                    ele_str = escape_html_chars(seg_data.get_value(ref_des))
                    if i in list(ele_pos_map.keys()) and ele_pos_map[i] == j:
                        ele_str = self._wrap_ele_error(ele_str)
                    t_seg[-1].append(ele_str)
            else:
                ref_des = '%02i' % (i)
                ele_str = escape_html_chars(seg_data.get_value(ref_des))
                if i in list(ele_pos_map.keys()):
                    ele_str = self._wrap_ele_error(ele_str)
                t_seg.append(ele_str)

        for err_node in err_node_list:
            #for err_tuple in err_node.errors:
            for err_tuple in err_node.get_error_list(seg_data.get_seg_id(), True):
                err_cde = err_tuple[0]
                err_str = err_tuple[1]
                if err_cde == '3':
                    self.fd.write('<span class="error">&nbsp;%s (Segment Error Code: %s)</span><br />\n' %
                                  (err_str, err_cde))
        if self.loop_info:
            self.gen_info(self.loop_info)
        self.loop_info = None
        self.fd.write('<span class="seg">%i:&nbsp;%s</span><br />\n' %
                      (cur_line, self._seg_str(seg_data.get_seg_id(), t_seg)))
        for err_node in err_node_list:
            for err_tuple in err_node.get_error_list(seg_data.get_seg_id(), False):
            #for err_tuple in err_node.errors:
                err_cde = err_tuple[0]
                err_str = err_tuple[1]
                if err_cde != '3':
                    self.fd.write('<span class="error">&nbsp;%s (Segment Error Code: %s)</span><br />\n' %
                                  (err_str, err_cde))
            for ele in err_node.elements:
                for (err_cde, err_str, err_val) in ele.get_error_list(seg_data.get_seg_id(), False):
                #for (err_cde, err_str, err_val) in ele.errors:
                    if not (seg_data.get_seg_id() == 'GE' and 'GS' in err_str):  # Ugly hack
                        self.fd.write('<span class="error">&nbsp;%s (Element Error Code: %s)</span><br />\n' %
                                      (err_str, err_cde))

    def _seg_str(self, seg_id, ele_list):
        """
        @param ele_list: list of formatted elements
        @rtype: string
        """
        return seg_id + self.ele_term + seg_str(
            ele_list, self.seg_term, self.ele_term,
            self.subele_term, self.eol)

    def _wrap_ele_error(self, str1):
        """
        @rtype: string
        """
        return '<span class="ele_err">%s</span>' % (str1)


def seg_str(seg, seg_term, ele_term, subele_term, eol=''):
    """
    Join a list of elements
    @param seg: List of elements
    @type seg: list[string|list[string]]
    @param seg_term: Segment terminator character
    @type seg_term: string
    @param ele_term: Element terminator character
    @type ele_term: string
    @param subele_term: Sub-element terminator character
    @type subele_term: string
    @param eol: End of line character
    @type eol: string
    @return: formatted segment
    @rtype: string
    """
    #if None in seg:
    #    logger.debug(seg)
    tmp = []
    for a in seg:
        if type(a) is ListType:
            tmp.append(subele_term.join(a))
        else:
            tmp.append(a)
    return '%s%s%s' % (ele_term.join(tmp), seg_term, eol)


def escape_html_chars(str_val):
    """
    Escape special HTML characters (& <>)
    @type str_val: string
    @return: formatted string
    @rtype: string
    """
    if str_val is None:
        return None
    output = str_val
    output = output.replace('&', '&amp;')
    output = output.replace(' ', '&nbsp;')
    output = output.replace('>', '&gt;')
    output = output.replace('<', '&lt;')
    return output



######################################################################
# Copyright (c)
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""pyx12 package exception classes.
"""


class XML_Reader_Error(Exception):
    """Class for XML Reader errors."""


class X12Error(Exception):
    """Base class for X12N format errors."""

#class ISAError(X12Error):
#    """Errors in the ISA or IEA segements."""


class GSError(X12Error):
    """Errors in the GS or GE segements."""

#class STError(X12Error):
#    """Errors in the ST or SE segements."""

#class WEDIError(Exception):
#    """Base class for WEDI errors in this module."""

#class WEDI1Error(WEDIError): pass
#class WEDI2Error(WEDIError): pass
#class WEDI3Error(WEDIError): pass
#class WEDI4Error(WEDIError): pass
#class WEDI5Error(WEDIError): pass
#class WEDI6Error(WEDIError): pass


class EngineError(Exception):
    """Base class for translation engine errors."""

#class HL_Loop_Pop(EngineError):
#    """Pop a HL level"""


class IterOutOfBounds(Exception):
    """Iterator is out of bounds"""


class IterDone(Exception):
    """Iterator is Complete"""


class IsValidError(Exception):
    """
    Exception for invalid X12 type errors
    """
    pass


class X12PathError(Exception):
    """
    Exception for invalid X12 path errors
    """
    pass



######################################################################
# Copyright (c) 2001-2013 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Walk a tree of x12_map nodes.  Find the correct node.

If seg indicates a loop has been entered, returns the first child segment node.
If seg indicates a segment has been entered, returns the segment node.
"""

import logging

# Intrapackage imports
from errors import EngineError
import pyx12.segment
from nodeCounter import NodeCounter

logger = logging.getLogger('pyx12.walk_tree')
#logger.setLevel(logging.DEBUG)
#logger.setLevel(logging.ERROR)


def pop_to_parent_loop(node):
    """
    @param node: Loop Node
    @type node: L{node<map_if.x12_node>}
    @return: Closest parent loop node
    @rtype: L{node<map_if.x12_node>}
    """
    if node.is_map_root():
        return node
    map_node = node.parent
    if map_node is None:
        raise EngineError("Node is None: %s" % (node.name))
    while not (map_node.is_loop() or map_node.is_map_root()):
        map_node = map_node.parent
    if not (map_node.is_loop() or map_node.is_map_root()):
        raise EngineError("Called pop_to_parent_loop, can't find parent loop")
    return map_node


def is_first_seg_match2(child, seg_data):
    """
    Find the first segment in loop, verify it matches segment

    @param child: child node
    @type child: L{node<map_if.x12_node>}
    @param seg_data: Segment object
    @type seg_data: L{segment<segment.Segment>}
    @rtype: boolean
    """
    if child.is_segment():
        if child.is_match(seg_data):
            return True
        else:
            # seg does not match the first segment in loop, so not valid
            return False
    return False


def get_id_list(node_list):
    # get_id_list(pop)
    ret = []
    for node in node_list:
        if node is not None:
            ret.append(node.id)
    return ret


def traverse_path(start_node, pop_loops, push_loops):
    """
    Debug function - From the start path, pop up then push down to get a path string
    """
    start_path = pop_to_parent_loop(start_node).get_path()
    p1 = [p for p in start_path.split('/') if p != '']
    for loop_id in get_id_list(pop_loops):
        assert loop_id == p1[-1], 'Path %s does not contain %s' % (start_path, loop_id)
        p1 = p1[:-1]
    for loop_id in get_id_list(push_loops):
        p1.append(loop_id)
    return '/' + '/'.join(p1)


class walk_tree(object):
    """
    Walks a map_if tree.  Tracks loop/segment counting, missing loop/segment.
    """
    def __init__(self, initialCounts=None):
        # Store errors until we know we have an error
        self.mandatory_segs_missing = []
        if initialCounts is None:
            initialCounts = {}
        self.counter = NodeCounter(initialCounts)

    def walk(self, node, seg_data, errh, seg_count, cur_line, ls_id):
        """
        Walk the node tree from the starting node to the node matching
        seg_data. Catch any counting or requirement errors along the way.

        Handle required segment/loop missed (not found in seg)
        Handle found segment = Not used

        @param node: Starting node
        @type node: L{node<map_if.x12_node>}
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        @param seg_count: Count of current segment in the ST Loop
        @type seg_count: int
        @param cur_line: Current line number in the file
        @type cur_line: int
        @param ls_id: The current LS loop identifier
        @type ls_id: string
        @return: The matching x12 segment node, a list of x12 popped loops, and a list
            of x12 pushed loops from the start segment to the found segment
        @rtype: (L{node<map_if.segment_if>}, [L{node<map_if.loop_if>}], [L{node<map_if.loop_if>}])

        @todo: check single segment loop repeat
        """
        pop_node_list = []
        push_node_list = []
        orig_node = node
        #logger.info('%s seg_count=%i / cur_line=%i' % (node.id, seg_count, cur_line))
        self.mandatory_segs_missing = []
        node_pos = node.pos  # Get original position ordinal of starting node
        if not (node.is_loop() or node.is_map_root()):
            node = pop_to_parent_loop(node)  # Get enclosing loop
            #node_list.append(node)
        while True:
            # Iterate through nodes with position >= current position
            for ord1 in [a for a in sorted(node.pos_map) if a >= node_pos]:
                for child in node.pos_map[ord1]:
                    if child.is_segment():
                        if child.is_match(seg_data):
                            # Is the matched segment the beginning of a loop?
                            if node.is_loop() \
                                    and self._is_loop_match(node, seg_data, errh, seg_count, cur_line, ls_id):
                                (
                                    node1, push_node_list) = self._goto_seg_match(node, seg_data,
                                                                                  errh, seg_count, cur_line, ls_id)
                                if orig_node.is_loop() or orig_node.is_map_root():
                                    orig_loop = orig_node
                                else:
                                    orig_loop = pop_to_parent_loop(orig_node)  # Get enclosing loop
                                if node == orig_loop:
                                    pop_node_list = [node]
                                    push_node_list = [node]
                                return (node1, pop_node_list, push_node_list)  # segment node
                            #child.incr_cur_count()
                            self.counter.increment(child.x12path)
                            #assert child.get_cur_count() == self.counter.get_count(child.x12path), \
                            #    'child counts not equal: old is %s=%i : new is %s=%i' % (
                            #    child.get_path(), child.get_cur_count(),
                            #    child.x12path.format(), self.counter.get_count(child.x12path))
                            self._check_seg_usage(child, seg_data, seg_count, cur_line, ls_id, errh)
                            # Remove any previously missing errors for this segment
                            self.mandatory_segs_missing = [x for x in self.mandatory_segs_missing if x[0] != child]
                            self._flush_mandatory_segs(errh, child.pos)
                            return (child, pop_node_list, push_node_list)  # segment node
                        elif child.usage == 'R' and self.counter.get_count(child.x12path) < 1:
                            fake_seg = pyx12.segment.Segment('%s' % (child.id), '~', '*', ':')
                            err_str = 'Mandatory segment "%s" (%s) missing' % (child.name, child.id)
                            self.mandatory_segs_missing.append((child, fake_seg, '3', err_str, seg_count, cur_line, ls_id))
                        #else:
                            #logger.debug('Segment %s is not a match for (%s*%s)' % \
                            #   (child.id, seg_data.get_seg_id(), seg_data[0].get_value()))
                    elif child.is_loop():
                        if self._is_loop_match(child, seg_data, errh, seg_count, cur_line, ls_id):
                            (node_seg, push_node_list) = self._goto_seg_match(child, seg_data, errh, seg_count, cur_line, ls_id)
                            return (node_seg, pop_node_list, push_node_list)  # segment node
            # End for ord1 in pos_keys
            if node.is_map_root():  # If at root and we haven't found the segment yet.
                walk_tree._seg_not_found_error(orig_node, seg_data,
                                               errh, seg_count, cur_line, ls_id)
                return (None, [], [])
            node_pos = node.pos  # Get position ordinal of current node in tree
            pop_node_list.append(node)
            node = pop_to_parent_loop(node)  # Get enclosing parent loop

        walk_tree._seg_not_found_error(orig_node, seg_data, errh, seg_count, cur_line, ls_id)
        return (None, [], [])

    def getCountState(self):
        return self.counter.getState()

    def setCountState(self, initialCounts={}):
        self.counter = NodeCounter(initialCounts)

    def forceWalkCounterToLoopStart(self, x12_path, child_path):
        # delete child counts under the x12_path, no longer needed
        self.counter.reset_to_node(x12_path)
        self.counter.increment(x12_path)  # add a count for this path
        self.counter.increment(child_path) # count the loop start segment

    def _check_seg_usage(self, seg_node, seg_data, seg_count, cur_line, ls_id, errh):
        """
        Check segment usage requirement and count

        @param seg_node: Segment X12 node to verify
        @type seg_node: L{node<map_if.segment_if>}
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        @param seg_count: Count of current segment in the ST Loop
        @type seg_count: int
        @param cur_line: Current line number in the file
        @type cur_line: int
        @param ls_id: The current LS loop identifier
        @type ls_id: string
        @param errh: Error handler
        @type errh: L{error_handler.err_handler}
        @raise EngineError: On invalid usage code
        """
        assert seg_node.usage in ('N', 'R', 'S'), 'Segment usage must be R, S, or N'
        if seg_node.usage == 'N':
            err_str = "Segment %s found but marked as not used" % (seg_node.id)
            errh.seg_error('2', err_str, None)
        elif seg_node.usage == 'R' or seg_node.usage == 'S':
            #assert seg_node.get_cur_count() == self.counter.get_count(seg_node.x12path), 'seg_node counts not equal'
            if self.counter.get_count(seg_node.x12path) > seg_node.get_max_repeat():  # handle seg repeat count
                err_str = "Segment %s exceeded max count.  Found %i, should have %i" \
                    % (seg_data.get_seg_id(), self.counter.get_count(seg_node.x12path), seg_node.get_max_repeat())
                errh.add_seg(seg_node, seg_data, seg_count, cur_line, ls_id)
                errh.seg_error('5', err_str, None)

    @staticmethod
    def _seg_not_found_error(orig_node, seg_data, errh, seg_count, cur_line, ls_id):
        """
        Create error for not found segments

        @param orig_node: Original starting node
        @type orig_node: L{node<map_if.x12_node>}
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        @param errh: Error handler
        @type errh: L{error_handler.err_handler}
        """
        if seg_data.get_seg_id() == 'HL':
            seg_str = seg_data.format('', '*', ':')
        else:
            seg_str = '%s*%s' % (seg_data.get_seg_id(), seg_data.get_value('01'))
        err_str = 'Segment %s not found.  Started at %s' % (seg_str, orig_node.get_path())
        errh.add_seg(orig_node, seg_data, seg_count, cur_line, ls_id)
        errh.seg_error('1', err_str, None)

    def _flush_mandatory_segs(self, errh, cur_pos=None):
        """
        Handle error reporting for any outstanding missing mandatory segments

        @param errh: Error handler
        @type errh: L{error_handler.err_handler}
        """
        for (seg_node, seg_data, err_cde, err_str, seg_count, cur_line, ls_id) in self.mandatory_segs_missing:
            # Create errors if not also at current position
            if seg_node.pos != cur_pos:
                errh.add_seg(seg_node, seg_data, seg_count, cur_line, ls_id)
                errh.seg_error(err_cde, err_str, None)
        self.mandatory_segs_missing = [x for x in self.mandatory_segs_missing if x[0].pos == cur_pos]

    def _is_loop_match(self, loop_node, seg_data, errh, seg_count, cur_line, ls_id):
        """
        Try to match the current loop to the segment
        Handle loop and segment counting.
        Check for used/missing

        @param loop_node: Loop Node
        @type loop_node: L{node<map_if.loop_if>}
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        @param errh: Error handler
        @type errh: L{error_handler.err_handler}

        @return: Does the segment match the first segment node in the loop?
        @rtype: boolean
        """
        assert loop_node.is_loop(), "Call to first_seg_match failed, node %s is not a loop. seg %s" \
            % (loop_node.id, seg_data.get_seg_id())
        #if loop_node.id not in ('ISA_LOOP', 'GS_LOOP'):
        #    assert loop_node.get_cur_count() == self.counter.get_count(loop_node.x12path), \
        #        'loop_node counts not equal: old is %s=%i : new is %s=%i' % (
        #        loop_node.get_path(), loop_node.get_cur_count(),
        #        loop_node.x12path.format(), self.counter.get_count(loop_node.x12path))
        if len(loop_node) <= 0:  # Has no children
            return False
        first_child_node = loop_node.get_first_node()
        assert first_child_node is not None, 'get_first_node failed from loop %s' % (loop_node.id)
        if first_child_node.is_loop():
            #If any loop node matches
            for child_node in loop_node.childIterator():
                if child_node.is_loop() and self._is_loop_match(child_node,
                                                                seg_data, errh, seg_count, cur_line, ls_id):
                    return True
        elif is_first_seg_match2(first_child_node, seg_data):
            return True
        elif loop_node.usage == 'R' and self.counter.get_count(loop_node.x12path) < 1:
            fake_seg = pyx12.segment.Segment('%s' % (first_child_node.id), '~', '*', ':')
            err_str = 'Mandatory loop "%s" (%s) missing' % \
                (loop_node.name, loop_node.id)
            self.mandatory_segs_missing.append((first_child_node, fake_seg,
                                                '3', err_str, seg_count, cur_line, ls_id))
        return False

    def _goto_seg_match(self, loop_node, seg_data, errh, seg_count, cur_line, ls_id):
        """
        A child loop has matched the segment.  Return that segment node.
        Handle loop counting and requirement errors.

        @param loop_node: The starting loop node.
        @type loop_node: L{node<map_if.loop_if>}
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        @param errh: Error handler
        @type errh: L{error_handler.err_handler}
        @param seg_count: Current segment count for ST loop
        @type seg_count: int
        @param cur_line: File line counter
        @type cur_line: int
        @type ls_id: string

        @return: The matching segment node and a list of the push loop nodes
        @rtype: (L{node<map_if.segment_if>}, [L{node<map_if.loop_if>}])
        """
        assert loop_node.is_loop(), "_goto_seg_match failed, node %s is not a loop. seg %s" \
            % (loop_node.id, seg_data.get_seg_id())
        first_child_node = loop_node.get_first_seg()
        if first_child_node is not None and is_first_seg_match2(first_child_node, seg_data):
            self._check_loop_usage(loop_node, seg_data,
                                   seg_count, cur_line, ls_id, errh)
            #first_child_node.incr_cur_count()
            self.counter.increment(first_child_node.x12path)
            #assert first_child_node.get_cur_count() == self.counter.get_count(first_child_node.x12path), 'first_child_node counts not equal'
            self._flush_mandatory_segs(errh)
            return (first_child_node, [loop_node])
        else:
            for child in loop_node.childIterator():
                if child.is_loop():
                    (
                        node1, push1) = self._goto_seg_match(child, seg_data, errh,
                                                             seg_count, cur_line, ls_id)
                    if node1:
                        push_node_list = [loop_node]
                        push_node_list.extend(push1)
                        return (node1, push_node_list)
        return (None, [])

    def _check_loop_usage(self, loop_node, seg_data, seg_count, cur_line, ls_id, errh):
        """
        Check loop usage requirement and count

        @param loop_node: Loop X12 node to verify
        @type loop_node: L{node<map_if.loop_if>}
        @param seg_data: Segment object
        @type seg_data: L{segment<segment.Segment>}
        @param seg_count: Count of current segment in the ST Loop
        @type seg_count: int
        @param cur_line: Current line number in the file
        @type cur_line: int
        @param ls_id: The current LS loop identifier
        @type ls_id: string
        @param errh: Error handler
        @type errh: L{error_handler.err_handler}
        @raise EngineError: On invalid usage code
        """
        assert loop_node.is_loop(), "Node %s is not a loop. seg %s" % (
            loop_node.id, seg_data.get_seg_id())
        assert loop_node.usage in ('N', 'R', 'S'), 'Loop usage must be R, S, or N'
        if loop_node.usage == 'N':
            err_str = "Loop %s found but marked as not used" % (loop_node.id)
            errh.seg_error('2', err_str, None)
        elif loop_node.usage in ('R', 'S'):
            #if loop_node.id == '2110':
            #    import ipdb; ipdb.set_trace()
            #loop_node.reset_child_count()
            self.counter.reset_to_node(loop_node.x12path)
            #loop_node.incr_cur_count()
            self.counter.increment(loop_node.x12path)
            #assert loop_node.get_cur_count() == self.counter.get_count(loop_node.x12path), \
            #    'loop_node counts not equal: old is %s=%i : new is %s=%i' % (
            #    loop_node.get_path(), loop_node.get_cur_count(),
            #    loop_node.x12path.format(), self.counter.get_count(loop_node.x12path))
            #logger.debug('incr loop_node %s %i' % (loop_node.id, loop_node.cur_count))
            #logger.debug('incr first_child_node %s %i' % (first_child_node.id, first_child_node.cur_count))
            if self.counter.get_count(loop_node.x12path) > loop_node.get_max_repeat():
                err_str = "Loop %s exceeded max count.  Found %i, should have %i" \
                    % (loop_node.id, self.counter.get_count(loop_node.x12path), loop_node.get_max_repeat())
                errh.add_seg(loop_node, seg_data, seg_count, cur_line, ls_id)
                errh.seg_error('4', err_str, None)
            #logger.debug('MATCH Loop %s / Segment %s (%s*%s)' \
            #    % (child.id, first_child_node.id, seg_data.get_seg_id(), seg[0].get_value()))



#####################################################################
# Copyright 
#   John Holland <john@zoner.org>
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Parse a ANSI X12N data file.  Validate against a map and codeset values.
Create XML, HTML, and 997/999 documents based on the data file.
"""

import logging

# Intrapackage imports
import pyx12.error_handler
import pyx12.error_997
import pyx12.error_999
import pyx12.error_html
import pyx12.errors
import pyx12.map_index
import pyx12.map_if
import pyx12.x12file
from pyx12.map_walker import walk_tree
import pyx12.x12xml_simple


def _reset_counter_to_isa_counts(walker):
    """
    Reset ISA instance counts
    """
    walker.counter.reset_to_node('/ISA_LOOP')
    walker.counter.increment('/ISA_LOOP')
    walker.counter.increment('/ISA_LOOP/ISA')


def _reset_counter_to_gs_counts(walker):
    """
    Reset GS instance counts
    """
    walker.counter.reset_to_node('/ISA_LOOP/GS_LOOP')
    walker.counter.increment('/ISA_LOOP/GS_LOOP')
    walker.counter.increment('/ISA_LOOP/GS_LOOP/GS')


def x12n_document(param, src_file, fd_997, fd_html,
                  fd_xmldoc=None, xslt_files=None, map_path=None):
    """
    Primary X12 validation function
    @param param: pyx12.param instance
    @param src_file: Source document
    @type src_file: string
    @param fd_997: 997/999 output document
    @type fd_997: file descriptor
    @param fd_html: HTML output document
    @type fd_html: file descriptor
    @param fd_xmldoc: XML output document
    @type fd_xmldoc: file descriptor
    @rtype: boolean
    """
    logger = logging.getLogger('pyx12')
    errh = pyx12.error_handler.err_handler()

    # Get X12 DATA file
    try:
        src = pyx12.x12file.X12Reader(src_file)
    except pyx12.errors.X12Error:
        logger.error('"%s" does not look like an X12 data file' % (src_file))
        return False

    #Get Map of Control Segments
    map_file = 'x12.control.00501.xml' if src.icvn == '00501' else 'x12.control.00401.xml'
    logger.debug('X12 control file: %s' % (map_file))
    control_map = pyx12.map_if.load_map_file(map_file, param, map_path)
    map_index_if = pyx12.map_index.map_index(map_path)
    node = control_map.getnodebypath('/ISA_LOOP/ISA')
    walker = walk_tree()
    icvn = fic = vriic = tspc = None
    cur_map = None  # we do not initially know the X12 transaction type
    #XXX Generate TA1 if needed.

    if fd_html:
        html = pyx12.error_html.error_html(errh, fd_html, src.get_term())
        html.header()
        err_iter = pyx12.error_handler.err_iter(errh)
    if fd_xmldoc:
        xmldoc = pyx12.x12xml_simple.x12xml_simple(fd_xmldoc, param.get('simple_dtd'))

    #basedir = os.path.dirname(src_file)
    #erx = errh_xml.err_handler(basedir=basedir)

    valid = True
    for seg in src:
        #find node
        orig_node = node

        if False:
            print('--------------------------------------------')
            print(seg)
            print('--------------------------------------------')
            # reset to control map for ISA and GS loops
            print('------- counters before --------')
            print(walker.counter._dict)
        if seg.get_seg_id() == 'ISA':
            node = control_map.getnodebypath('/ISA_LOOP/ISA')
            walker.forceWalkCounterToLoopStart('/ISA_LOOP', '/ISA_LOOP/ISA')
        elif seg.get_seg_id() == 'GS':
            node = control_map.getnodebypath('/ISA_LOOP/GS_LOOP/GS')
            walker.forceWalkCounterToLoopStart('/ISA_LOOP/GS_LOOP', '/ISA_LOOP/GS_LOOP/GS')
        else:
            # from the current node, find the map node matching the segment
            # keep track of the loops traversed
            try:
                (node, pop_loops, push_loops) = walker.walk(node, seg, errh,
                    src.get_seg_count(), src.get_cur_line(), src.get_ls_id())
            except pyx12.errors.EngineError:
                logger.error('Source file line %i' % (src.get_cur_line()))
                raise

        if False:
            print('------- counters after --------')
            print(walker.counter._dict)
        if node is None:
            node = orig_node
        else:
            if seg.get_seg_id() == 'ISA':
                errh.add_isa_loop(seg, src)
                icvn = seg.get_value('ISA12')
                errh.handle_errors(src.pop_errors())
            elif seg.get_seg_id() == 'IEA':
                errh.handle_errors(src.pop_errors())
                errh.close_isa_loop(node, seg, src)
                # Generate 997
                #XXX Generate TA1 if needed.
            elif seg.get_seg_id() == 'GS':
                fic = seg.get_value('GS01')
                vriic = seg.get_value('GS08')
                map_file_new = map_index_if.get_filename(icvn, vriic, fic)
                if map_file != map_file_new:
                    map_file = map_file_new
                    if map_file is None:
                        err_str = "Map not found.  icvn={}, fic={}, vriic={}".format(icvn, fic, vriic)
                        raise pyx12.errors.EngineError(err_str)
                    cur_map = pyx12.map_if.load_map_file(map_file, param, map_path)
                    src.check_837_lx = True if cur_map.id == '837' else False
                    logger.debug('Map file: %s' % (map_file))
                    #apply_loop_count(orig_node, cur_map)
                    #reset_isa_counts(cur_map)
                    #_reset_counter_to_isa_counts(walker)  # new counter
                #reset_gs_counts(cur_map)
                #_reset_counter_to_gs_counts(walker)  # new counter
                node = cur_map.getnodebypath('/ISA_LOOP/GS_LOOP/GS')
                errh.add_gs_loop(seg, src)
                errh.handle_errors(src.pop_errors())
            elif seg.get_seg_id() == 'BHT':
                # special case for 4010 837P
                if vriic in ('004010X094', '004010X094A1'):
                    tspc = seg.get_value('BHT02')
                    logger.debug('icvn=%s, fic=%s, vriic=%s, tspc=%s' %
                                 (icvn, fic, vriic, tspc))
                    map_file_new = map_index_if.get_filename(icvn, vriic, fic, tspc)
                    logger.debug('New map file: %s' % (map_file_new))
                    if map_file != map_file_new:
                        map_file = map_file_new
                        if map_file is None:
                            err_str = "Map not found.  icvn={}, fic={}, vriic={}, tspc={}".format(
                                        icvn, fic, vriic, tspc)
                            raise pyx12.errors.EngineError(err_str)
                        cur_map = pyx12.map_if.load_map_file(map_file, param, map_path)
                        src.check_837_lx = True if cur_map.id == '837' else False
                        logger.debug('Map file: %s' % (map_file))
                        #apply_loop_count(node, cur_map)
                        node = cur_map.getnodebypath('/ISA_LOOP/GS_LOOP/ST_LOOP/HEADER/BHT')
                errh.add_seg(node, seg, src.get_seg_count(), src.get_cur_line(), src.get_ls_id())
                errh.handle_errors(src.pop_errors())
            elif seg.get_seg_id() == 'GE':
                errh.handle_errors(src.pop_errors())
                errh.close_gs_loop(node, seg, src)
            elif seg.get_seg_id() == 'ST':
                errh.add_st_loop(seg, src)
                errh.handle_errors(src.pop_errors())
            elif seg.get_seg_id() == 'SE':
                errh.handle_errors(src.pop_errors())
                errh.close_st_loop(node, seg, src)
            else:
                errh.add_seg(node, seg, src.get_seg_count(), src.get_cur_line(), src.get_ls_id())
                errh.handle_errors(src.pop_errors())

            #errh.set_cur_line(src.get_cur_line())
            valid &= node.is_valid(seg, errh)
            #erx.handleErrors(src.pop_errors())
            #erx.handleErrors(errh.get_errors())
            #errh.reset()

        if fd_html:
            if node is not None and node.is_first_seg_in_loop():
                html.loop(node.get_parent())
            err_node_list = []
            while True:
                try:
                    err_iter.next()
                    err_node = err_iter.get_cur_node()
                    err_node_list.append(err_node)
                except pyx12.errors.IterOutOfBounds:
                    break
            html.gen_seg(seg, src, err_node_list)

        if fd_xmldoc:
            xmldoc.seg(node, seg)

        if False:
            print('\n\n')
        #erx.Write(src.cur_line)

    #erx.handleErrors(src.pop_errors())
    src.cleanup()  # Catch any skipped loop trailers
    errh.handle_errors(src.pop_errors())
    #erx.handleErrors(src.pop_errors())
    #erx.handleErrors(errh.get_errors())

    if fd_html:
        html.footer()
        del html

    if fd_xmldoc:
        del xmldoc

    #visit_debug = pyx12.error_debug.error_debug_visitor(sys.stdout)
    #errh.accept(visit_debug)

    #If this transaction is not a 997/999, generate one.
    if fd_997 and fic != 'FA':
        if vriic and vriic[:6] == '004010':
            try:
                visit_997 = pyx12.error_997.error_997_visitor(fd_997, src.get_term())
                errh.accept(visit_997)
                del visit_997
            except Exception:
                logger.exception('Failed to create 997 response')
        if vriic and vriic[:6] == '005010':
            try:
                visit_999 = pyx12.error_999.error_999_visitor(fd_997, src.get_term())
                errh.accept(visit_999)
                del visit_999
            except Exception:
                logger.exception('Failed to create 999 response')
    del node
    del src
    del control_map
    try:
        del cur_map
    except UnboundLocalError:
        pass
    try:
        if not valid or errh.get_error_count() > 0:
            return False
        else:
            return True
    except Exception:
        print(errh)
        return False



######################################################################
# Copyright (c) 
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.txt, which
# you should have received as part of this distribution.
#
######################################################################

"""
Interface to a X12N IG Map
"""
import logging
import os.path
import string
import sys
import re
import xml.etree.cElementTree as et
from pkg_resources import resource_stream

# Intrapackage imports
from errors import EngineError
import codes
import dataele
import path
import validation
from syntax import is_syntax_valid

MAXINT = 2147483647


class x12_node(object):
    """
    X12 Node Superclass
    """
    def __init__(self):
        self.id = None
        self.name = None
        self.parent = None
        self.children = []
        self.path = ''
        self._x12path = None
        self._fullpath = None

    def __eq__(self, other):
        if isinstance(other, x12_node):
            return self.id == other.id and self.parent.id == other.parent.id
        return NotImplemented

    def __ne__(self, other):
        res = type(self).__eq__(self, other)
        if res is NotImplemented:
            return res
        return not res

    def __lt__(self, other):
        return NotImplemented

    __le__ = __lt__
    __le__ = __lt__
    __gt__ = __lt__
    __ge__ = __lt__

    def __hash__(self):
        return (self.id + self.parent.id).__hash__()

    def __len__(self):
        return len(self.children)

    def __repr__(self):
        """
        @rtype: string
        """
        return self.name

    def getnodebypath(self, path):
        """
        """
        pathl = path.split('/')
        if len(pathl) == 0:
            return None
        for child in self.children:
            if child.id.lower() == pathl[0].lower():
                if len(pathl) == 1:
                    return child
                else:
                    if child.is_loop():
                        return child.getnodebypath(string.join(pathl[1:], '/'))
                    else:
                        break
        raise EngineError('getnodebypath failed. Path "%s" not found' % path)

    def get_child_count(self):
        return len(self.children)

    def get_child_node_by_idx(self, idx):
        """
        @param idx: zero based
        """
        if idx >= len(self.children):
            return None
        else:
            return self.children[idx]

    def get_path(self):
        """
        @return: path - XPath style
        @rtype: string
        """
        if self._fullpath:
            return self._fullpath
        parent_path = self.parent.get_path()
        if parent_path == '/':
            self._fullpath = '/' + self.path
            return self._fullpath
        else:
            self._fullpath = parent_path + '/' + self.path
            return self._fullpath

    def _get_x12_path(self):
        """
        @return: X12 node path
        @rtype: L{path<path.X12Path>}
        """
        if self._x12path:
            return self._x12path
        p = path.X12Path(self.get_path())
        self._x12path = p
        return p

    x12path = property(_get_x12_path, None, None)

    def is_first_seg_in_loop(self):
        """
        @rtype: boolean
        """
        return False

    def is_map_root(self):
        """
        @rtype: boolean
        """
        return False

    def is_loop(self):
        """
        @rtype: boolean
        """
        return False

    def is_segment(self):
        """
        @rtype: boolean
        """
        return False

    def is_element(self):
        """
        @rtype: boolean
        """
        return False

    def is_composite(self):
        """
        @rtype: boolean
        """
        return False


############################################################
# Map file interface
############################################################
class map_if(x12_node):
    """
    Map file interface
    """
    def __init__(self, eroot, param, base_path=None):
        """
        @param eroot: ElementTree root
        @param param: map of parameters
        """
        x12_node.__init__(self)
        self.children = None
        self.pos_map = {}
        self.cur_path = '/transaction'
        self.path = '/'
        #self.cur_iter_node = self
        self.param = param
        #global codes
        self.ext_codes = codes.ExternalCodes(base_path,
                                             param.get('exclude_external_codes'))
        self.data_elements = dataele.DataElements(base_path)

        self.id = eroot.get('xid')

        self.name = eroot.get('name') if eroot.get('name') else eroot.findtext('name')
        self.base_name = 'transaction'
        for e in eroot.findall('loop'):
            loop_node = loop_if(self, self, e)
            try:
                self.pos_map[loop_node.pos].append(loop_node)
            except KeyError:
                self.pos_map[loop_node.pos] = [loop_node]
        for e in eroot.findall('segment'):
            seg_node = segment_if(self, self, e)
            try:
                self.pos_map[seg_node.pos].append(seg_node)
            except KeyError:
                self.pos_map[seg_node.pos] = [seg_node]
        self.icvn = self._get_icvn()

    def _get_icvn(self):
        """
        Get the Interchange version of this map
        Map must have a first ISA segment
        ISA12
        """
        ipath = '/ISA_LOOP/ISA'
        try:
            node = self.getnodebypath(ipath).children[11]
            icvn = node.valid_codes[0]
            return icvn
        except Exception:
            return None

    def debug_print(self):
        sys.stdout.write(self.__repr__())
        for ord1 in sorted(self.pos_map):
            for node in self.pos_map[ord1]:
                node.debug_print()

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return (self.id).__hash__()

    def __len__(self):
        i = 0
        for ord1 in sorted(self.pos_map):
            i += len(self.pos_map[ord1])
        return i

    def get_child_count(self):
        return self.__len__()

    def get_first_node(self):
        pos_keys = sorted(self.pos_map)
        if len(pos_keys) > 0:
            return self.pos_map[pos_keys[0]][0]
        else:
            return None

    def get_first_seg(self):
        first = self.get_first_node()
        if first.is_segment():
            return first
        else:
            return None

    def __repr__(self):
        """
        @rtype: string
        """
        return '%s\n' % (self.id)

    def _path_parent(self):
        """
        @rtype: string
        """
        return os.path.basename(os.path.dirname(self.cur_path))

    def get_path(self):
        """
        @rtype: string
        """
        return self.path

    def get_child_node_by_idx(self, idx):
        """
        @param idx: zero based
        """
        raise EngineError('map_if.get_child_node_by_idx is not a valid call')

    def getnodebypath(self, spath):
        """
        @param spath: Path string; /1000/2000/2000A/NM102-3
        @type spath: string
        """
        pathl = spath.split('/')[1:]
        if len(pathl) == 0:
            return None
        #logger.debug('%s %s %s' % (self.base_name, self.id, pathl[1]))
        for ord1 in sorted(self.pos_map):
            for child in self.pos_map[ord1]:
                if child.id.lower() == pathl[0].lower():
                    if len(pathl) == 1:
                        return child
                    else:
                        return child.getnodebypath(string.join(pathl[1:], '/'))
        raise EngineError('getnodebypath failed. Path "%s" not found' % spath)

    def getnodebypath2(self, path_str):
        """
        @param path: Path string; /1000/2000/2000A/NM102-3
        @type path: string
        """
        x12path = path.X12Path(path_str)
        if x12path.empty():
            return None
        for ord1 in sorted(self.pos_map):
            for child in self.pos_map[ord1]:
                if child.id.upper() == x12path.loop_list[0]:
                    if len(x12path.loop_list) > 1:
                        return child
                    else:
                        del x12path.loop_list[0]
                        return child.getnodebypath(x12path.format())
        raise EngineError(
            'getnodebypath failed. Path "%s" not found' % path_str)

    def is_map_root(self):
        """
        @rtype: boolean
        """
        return True

    def reset_child_count(self):
        """
        Set cur_count of child nodes to zero
        """
        raise DeprecationWarning('Moved to nodeCounter')
        for ord1 in sorted(self.pos_map):
            for child in self.pos_map[ord1]:
                child.reset_cur_count()

    def reset_cur_count(self):
        """
        Set cur_count of child nodes to zero
        """
        raise DeprecationWarning('Moved to nodeCounter')
        self.reset_child_count()

    def __iter__(self):
        return self

    def loop_segment_iterator(self):
        yield self
        for ord1 in sorted(self.pos_map):
            for child in self.pos_map[ord1]:
                if child.is_loop() or child.is_segment():
                    for c in child.loop_segment_iterator():
                        yield c


############################################################
# Loop Interface
############################################################
class loop_if(x12_node):
    """
    Loop Interface
    """
    def __init__(self, root, parent, elem):
        """
        """
        x12_node.__init__(self)
        self.root = root
        self.parent = parent
        self.pos_map = {}
        #self.path = ''
        self.base_name = 'loop'
        #self.type = 'implicit'
        self._cur_count = 0

        self.id = elem.get('xid')
        self.path = self.id
        self.type = elem.get('type')

        self.name = elem.get(
            'name') if elem.get('name') else elem.findtext('name')
        self.usage = elem.get(
            'usage') if elem.get('usage') else elem.findtext('usage')
        self.pos = int(elem.get(
            'pos')) if elem.get('pos') else int(elem.findtext('pos'))
        self.repeat = elem.get('repeat') if elem.get(
            'repeat') else elem.findtext('repeat')

        for e in elem.findall('loop'):
            loop_node = loop_if(self.root, self, e)
            #if self.pos_map:
            #    assert loop_node.pos >= max(self.pos_map.keys()), 'Bad ordinal %s' % (loop_node)
            try:
                self.pos_map[loop_node.pos].append(loop_node)
            except KeyError:
                self.pos_map[loop_node.pos] = [loop_node]
        for e in elem.findall('segment'):
            seg_node = segment_if(self.root, self, e)
            #if self.pos_map:
            #    assert seg_node.pos >= max(self.pos_map.keys()), 'Bad ordinal %s' % (seg_node)
            try:
                self.pos_map[seg_node.pos].append(seg_node)
            except KeyError:
                self.pos_map[seg_node.pos] = [seg_node]

        # For the segments with duplicate ordinals, adjust the path to be unique
        for ord1 in sorted(self.pos_map):
            if len(self.pos_map[ord1]) > 1:
                for seg_node in [n for n in self.pos_map[ord1] if n.is_segment()]:
                    id_elem = seg_node.guess_unique_key_id_element()
                    if id_elem is not None:
                        seg_node.path = seg_node.path + '[' + id_elem.valid_codes[0] + ']'

    def debug_print(self):
        sys.stdout.write(self.__repr__())
        for ord1 in sorted(self.pos_map):
            for node in self.pos_map[ord1]:
                node.debug_print()

    def __len__(self):
        i = 0
        for ord1 in sorted(self.pos_map):
            i += len(self.pos_map[ord1])
        return i

    def __repr__(self):
        """
        @rtype: string
        """
        out = ''
        if self.id:
            out += 'LOOP %s' % (self.id)
        if self.name:
            out += '  "%s"' % (self.name)
        if self.usage:
            out += '  usage: %s' % (self.usage)
        if self.pos:
            out += '  pos: %s' % (self.pos)
        if self.repeat:
            out += '  repeat: %s' % (self.repeat)
        out += '\n'
        return out

    def get_max_repeat(self):
        if self.repeat is None:
            return MAXINT
        if self.repeat == '&gt;1' or self.repeat == '>1':
            return MAXINT
        return int(self.repeat)

    def get_parent(self):
        return self.parent

    def get_first_node(self):
        pos_keys = sorted(self.pos_map)
        if len(pos_keys) > 0:
            return self.pos_map[pos_keys[0]][0]
        else:
            return None

    def get_first_seg(self):
        first = self.get_first_node()
        if first.is_segment():
            return first
        else:
            return None

    def childIterator(self):
        for ord1 in sorted(self.pos_map):
            for child in self.pos_map[ord1]:
                yield child

    def getnodebypath(self, spath):
        """
        @param spath: remaining path to match
        @type spath: string
        @return: matching node, or None is no match
        """
        pathl = spath.split('/')
        if len(pathl) == 0:
            return None
        for ord1 in sorted(self.pos_map):
            for child in self.pos_map[ord1]:
                if child.is_loop():
                    if child.id.upper() == pathl[0].upper():
                        if len(pathl) == 1:
                            return child
                        else:
                            return child.getnodebypath(string.join(pathl[1:], '/'))
                elif child.is_segment() and len(pathl) == 1:
                    if pathl[0].find('[') == -1:  # No id to match
                        if pathl[0] == child.id:
                            return child
                    else:
                        seg_id = pathl[0][0:pathl[0].find('[')]
                        id_val = pathl[0][pathl[0].find('[')
                                          + 1:pathl[0].find(']')]
                        if seg_id == child.id:
                            possible = child.get_unique_key_id_element(id_val)
                            if possible is not None:
                                return child
        raise EngineError('getnodebypath failed. Path "%s" not found' % spath)

    def getnodebypath2(self, path_str):
        """
        Try x12 path

        @param path_str: remaining path to match
        @type path_str: string
        @return: matching node, or None is no match
        """
        x12path = path.X12Path(path_str)
        if x12path.empty():
            return None
        for ord1 in sorted(self.pos_map):
            for child in self.pos_map[ord1]:
                if child.is_loop() and len(x12path.loop_list) > 0:
                    if child.id.upper() == x12path.loop_list[0].upper():
                        if len(x12path.loop_list) == 1 and x12path.seg_id is None:
                            return child
                        else:
                            return child.getnodebypath(x12path.format())
                elif child.is_segment() and len(x12path.loop_list) == 0 and x12path.seg_id is not None:
                    if x12path.id_val is None:
                        if x12path.seg_id == child.id:
                            return child
                    else:
                        seg_id = x12path.seg_id
                        id_val = x12path.id_val
                        if seg_id == child.id:
                            possible = child.get_unique_key_id_element(id_val)
                            if possible is not None:
                                return child
        raise EngineError(
            'getnodebypath failed. Path "%s" not found' % path_str)

    def get_child_count(self):
        return self.__len__()

    def get_child_node_by_idx(self, idx):
        """
        @param idx: zero based
        """
        raise EngineError('loop_if.get_child_node_by_idx is not a valid call for a loop_if')

    def get_seg_count(self):
        """
        @return: Number of child segments
        @rtype: integer
        """
        i = 0
        for ord1 in sorted(self.pos_map):
            for child in self.pos_map[ord1]:
                if child.is_segment():
                    i += 1
        return i

    def is_loop(self):
        """
        @rtype: boolean
        """
        return True

    def is_match(self, seg_data):
        """
        @type seg_data: L{segment<segment.Segment>}
        @return: Is the segment a match to this loop?
        @rtype: boolean
        """
        pos_keys = sorted(self.pos_map)
        child = self.pos_map[pos_keys[0]][0]
        if child.is_loop():
            return child.is_match(seg_data)
        elif child.is_segment():
            if child.is_match(seg_data):
                return True
            else:
                return False  # seg does not match the first segment in loop, so not valid
        else:
            return False

    def get_child_seg_node(self, seg_data):
        """
        Return the child segment matching the segment data
        """
        for child in self.childIterator():
            if child.is_segment() and child.is_match(seg_data):
                return child
        return None

    def get_child_loop_node(self, seg_data):
        """
        Return the child segment matching the segment data
        """
        for child in self.childIterator():
            if child.is_loop() and child.is_match(seg_data):
                return child
        return None

    def get_cur_count(self):
        """
        @return: current count
        @rtype: int
        """
        raise DeprecationWarning('Moved to nodeCounter')
        return self._cur_count

    def incr_cur_count(self):
        raise DeprecationWarning('Moved to nodeCounter')
        self._cur_count += 1

    def reset_child_count(self):
        """
        Set cur_count of child nodes to zero
        """
        raise DeprecationWarning('Moved to nodeCounter')
        for ord1 in sorted(self.pos_map):
            for child in self.pos_map[ord1]:
                child.reset_cur_count()

    def reset_cur_count(self):
        """
        Set cur_count of node and child nodes to zero
        """
        raise DeprecationWarning('Moved to nodeCounter')
        self._cur_count = 0
        self.reset_child_count()

    def set_cur_count(self, ct):
        raise DeprecationWarning('Moved to nodeCounter')
        self._cur_count = ct

    def get_counts_list(self, ct_list):
        """
        Build a list of (path, ct) of the current node and parents
        Gets the node counts to apply to another map
        @param ct_list: List to append to
        @type ct_list: list[(string, int)]
        """
        raise DeprecationWarning('Moved to nodeCounter')
        my_ct = (self.get_path(), self._cur_count)
        ct_list.append(my_ct)
        if not self.parent.is_map_root():
            self.parent.get_counts_list(ct_list)
        return True

    def loop_segment_iterator(self):
        yield self
        for ord1 in sorted(self.pos_map):
            for child in self.pos_map[ord1]:
                if child.is_loop() or child.is_segment():
                    for c in child.loop_segment_iterator():
                        yield c


class segment_if(x12_node):
    """
    Segment Interface
    """
    def __init__(self, root, parent, elem):
        """
        @param parent: parent node
        """

        x12_node.__init__(self)
        self.root = root
        self.parent = parent
        self.children = []
        #self.path = ''
        self.base_name = 'segment'
        self._cur_count = 0
        self.syntax = []

        self.id = elem.get('xid')
        self.path = self.id
        self.type = elem.get('type')

        self.name = elem.get(
            'name') if elem.get('name') else elem.findtext('name')
        self.usage = elem.get(
            'usage') if elem.get('usage') else elem.findtext('usage')
        self.pos = int(elem.get(
            'pos')) if elem.get('pos') else int(elem.findtext('pos'))
        self.max_use = elem.get('max_use') if elem.get(
            'max_use') else elem.findtext('max_use')
        self.repeat = elem.get('repeat') if elem.get(
            'repeat') else elem.findtext('repeat')

        self.end_tag = elem.get('end_tag') if elem.get(
            'end_tag') else elem.findtext('end_tag')

        for s in elem.findall('syntax'):
            syn_list = self._split_syntax(s.text)
            if syn_list is not None:
                self.syntax.append(syn_list)

        children_map = {}
        for e in elem.findall('element'):
            seq = int(e.get('seq')) if e.get('seq') else int(e.findtext('seq'))
            children_map[seq] = e
            #self.children.append(element_if(self.root, self, e))

        for e in elem.findall('composite'):
            seq = int(e.get('seq')) if e.get('seq') else int(e.findtext('seq'))
            children_map[seq] = e
            #self.children.append(composite_if(self.root, self, e))

        for seq in sorted(children_map.keys()):
            if children_map[seq].tag == 'element':
                self.children.append(element_if(
                    self.root, self, children_map[seq]))
            elif children_map[seq].tag == 'composite':
                self.children.append(composite_if(
                    self.root, self, children_map[seq]))

    def debug_print(self):
        sys.stdout.write(self.__repr__())
        for node in self.children:
            node.debug_print()

    def __repr__(self):
        """
        @rtype: string
        """
        out = '%s "%s"' % (self.id, self.name)
        if self.usage:
            out += '  usage: %s' % (self.usage)
        if self.pos:
            out += '  pos: %i' % (self.pos)
        if self.max_use:
            out += '  max_use: %s' % (self.max_use)
        out += '\n'
        return out

    def get_child_node_by_idx(self, idx):
        """
        @param idx: zero based
        """
        if idx >= len(self.children):
            return None
        else:
            m = [c for c in self.children if c.seq == idx + 1]
            if len(m) == 1:
                return m[0]
            else:
                raise EngineError('idx %i not found in %s' % (idx, self.id))

    def get_child_node_by_ordinal(self, ord):
        """
        Get a child element or composite by the X12 ordinal
        @param ord: one based element/composite index.  Corresponds to the map <seq> element
        @type ord: int
        """
        return self.get_child_node_by_idx(ord - 1)

    def get_max_repeat(self):
        if self.max_use is None or self.max_use == '>1':
            return MAXINT
        return int(self.max_use)

    def get_parent(self):
        """
        @return: ref to parent class instance
        @rtype: pyx12.x12_node
        """
        return self.parent

    def is_first_seg_in_loop(self):
        """
        @rtype: boolean
        """
        if self is self.get_parent().get_first_seg():
            return True
        else:
            return False

    def is_match(self, seg):
        """
        Is data segment given a match to this segment node?
        @param seg: data segment instance
        @return: boolean
        @rtype: boolean
        """
        if seg.get_seg_id() == self.id:
            if self.children[0].is_element() \
                    and self.children[0].get_data_type() == 'ID' \
                    and self.children[0].usage == 'R' \
                    and len(self.children[0].valid_codes) > 0 \
                    and seg.get_value('01') not in self.children[0].valid_codes:
                #logger.debug('is_match: %s %s' % (seg.get_seg_id(), seg[1]), self.children[0].valid_codes)
                return False
            # Special Case for 820
            elif seg.get_seg_id() == 'ENT' \
                    and self.children[1].is_element() \
                    and self.children[1].get_data_type() == 'ID' \
                    and len(self.children[1].valid_codes) > 0 \
                    and seg.get_value('02') not in self.children[1].valid_codes:
                #logger.debug('is_match: %s %s' % (seg.get_seg_id(), seg[1]), self.children[0].valid_codes)
                return False
            # Special Case for 999 CTX
            # IG defines the dataelement 2100/CT01-1 as an AN, but acts like an ID
            elif seg.get_seg_id() == 'CTX' \
                    and self.children[0].is_composite() \
                    and self.children[0].children[0].get_data_type() == 'AN' \
                    and len(self.children[0].children[0].valid_codes) > 0 \
                    and seg.get_value('01-1') not in self.children[0].children[0].valid_codes:
                return False
            elif self.children[0].is_composite() \
                    and self.children[0].children[0].get_data_type() == 'ID' \
                    and len(self.children[0].children[0].valid_codes) > 0 \
                    and seg.get_value('01-1') not in self.children[0].children[0].valid_codes:
                return False
            elif seg.get_seg_id() == 'HL' and self.children[2].is_element() \
                    and len(self.children[2].valid_codes) > 0 \
                    and seg.get_value('03') not in self.children[2].valid_codes:
                return False
            else:
                return True
        else:
            return False

    def is_match_qual(self, seg_data, seg_id, qual_code):
        """
        Is segment id and qualifier a match to this segment node and to this particulary segment data?
        @param seg_data: data segment instance
        @type seg_data: L{segment<segment.Segment>}
        @param seg_id: data segment ID
        @param qual_code: an ID qualifier code
        @return: True if a match
        @rtype: boolean
        """
        if seg_id == self.id:
            if qual_code is None:
                return True
            elif self.children[0].is_element() \
                    and self.children[0].get_data_type() == 'ID' \
                    and self.children[0].usage == 'R' \
                    and len(self.children[0].valid_codes) > 0:
                if qual_code in self.children[0].valid_codes and seg_data.get_value('01') == qual_code:
                    return True
                else:
                    return False
            # Special Case for 820
            elif seg_id == 'ENT' \
                    and self.children[1].is_element() \
                    and self.children[1].get_data_type() == 'ID' \
                    and len(self.children[1].valid_codes) > 0:
                if qual_code in self.children[1].valid_codes and seg_data.get_value('02') == qual_code:
                    return True
                else:
                    return False
            elif self.children[0].is_composite() \
                    and self.children[0].children[0].get_data_type() == 'ID' \
                    and len(self.children[0].children[0].valid_codes) > 0:
                if qual_code in self.children[0].children[0].valid_codes and seg_data.get_value('01-1') == qual_code:
                    return True
                else:
                    return False
            elif seg_id == 'HL' and self.children[2].is_element() \
                    and len(self.children[2].valid_codes) > 0:
                if qual_code in self.children[2].valid_codes and seg_data.get_value('03') == qual_code:
                    return True
                else:
                    return False
            else:
                return True
        else:
            return False

    def guess_unique_key_id_element(self):
        """
        Some segments, like REF, DTP, and DTP are duplicated.  They are matched using the value of an ID element.
        Which element to use varies.  This function tries to find a good candidate.
        """
        if self.children[0].is_element() and self.children[0].get_data_type() == 'ID' and len(self.children[0].valid_codes) > 0:
            return self.children[0]
        # Special Case for 820
        elif self.id == 'ENT' and self.children[1].is_element() and self.children[1].get_data_type() == 'ID' and len(self.children[1].valid_codes) > 0:
            return self.children[1]
        elif self.children[0].is_composite() and self.children[0].children[0].get_data_type() == 'ID' and len(self.children[0].children[0].valid_codes) > 0:
            return self.children[0].children[0]
        elif self.id == 'HL' and self.children[2].is_element() and len(self.children[2].valid_codes) > 0:
            return self.children[2]
        return None

    def get_unique_key_id_element(self, id_val):
        """
        Some segments, like REF, DTP, and DTP are duplicated.  They are matched using the value of an ID element.
        Which element to use varies.  This function tries to find a good candidate, using a key value
        """

        if self.children[0].is_element() and self.children[0].get_data_type() == 'ID' \
                and len(self.children[0].valid_codes) > 0 and id_val in self.children[0].valid_codes:
            return self.children[0]
        # Special Case for 820
        elif self.id == 'ENT' and self.children[1].is_element() and self.children[1].get_data_type() == 'ID' \
                and len(self.children[1].valid_codes) > 0 and id_val in self.children[1].valid_codes:
            return self.children[1]
        elif self.children[0].is_composite() and self.children[0].children[0].get_data_type() == 'ID' \
                and len(self.children[0].children[0].valid_codes) > 0 and id_val in self.children[0].children[0].valid_codes:
            return self.children[0].children[0]
        elif self.id == 'HL' and self.children[2].is_element() and len(self.children[2].valid_codes) > 0 and id_val in self.children[2].valid_codes:
            return self.children[2]
        return None

    def is_segment(self):
        """
        @rtype: boolean
        """
        return True

    def is_valid(self, seg_data, errh):
        """
        @param seg_data: data segment instance
        @type seg_data: L{segment<segment.Segment>}
        @param errh: instance of error_handler
        @rtype: boolean
        """
        valid = True
        child_count = self.get_child_count()
        if len(seg_data) > child_count:
            #child_node = self.get_child_node_by_idx(child_count+1)
            err_str = 'Too many elements in segment "%s" (%s). Has %i, should have %i' % \
                (self.name, seg_data.get_seg_id(), len(seg_data), child_count)
            #self.logger.error(err_str)
            ref_des = '%02i' % (child_count + 1)
            err_value = seg_data.get_value(ref_des)
            errh.ele_error('3', err_str, err_value, ref_des)
            valid = False

        dtype = []
        type_list = []
        for i in range(min(len(seg_data), child_count)):
            #self.logger.debug('i=%i, len(seg_data)=%i / child_count=%i' % \
            #   (i, len(seg_data), self.get_child_count()))
            child_node = self.get_child_node_by_idx(i)
            if child_node.is_composite():
                # Validate composite
                ref_des = '%02i' % (i + 1)
                comp_data = seg_data.get(ref_des)
                subele_count = child_node.get_child_count()
                if seg_data.ele_len(ref_des) > subele_count and child_node.usage != 'N':
                    subele_node = child_node.get_child_node_by_idx(
                        subele_count + 1)
                    err_str = 'Too many sub-elements in composite "%s" (%s)' % \
                        (subele_node.name, subele_node.refdes)
                    err_value = seg_data.get_value(ref_des)
                    errh.ele_error('3', err_str, err_value, ref_des)
                valid &= child_node.is_valid(comp_data, errh)
            elif child_node.is_element():
                # Validate Element
                if i == 1 and seg_data.get_seg_id() == 'DTP' \
                        and seg_data.get_value('02') in ('RD8', 'D8', 'D6', 'DT', 'TM'):
                    dtype = [seg_data.get_value('02')]
                if child_node.data_ele == '1250':
                    type_list.extend(child_node.valid_codes)
                ele_data = seg_data.get('%02i' % (i + 1))
                if i == 2 and seg_data.get_seg_id() == 'DTP':
                    valid &= child_node.is_valid(ele_data, errh, dtype)
                elif child_node.data_ele == '1251' and len(type_list) > 0:
                    valid &= child_node.is_valid(ele_data, errh, type_list)
                else:
                    valid &= child_node.is_valid(ele_data, errh)

        for i in range(min(len(seg_data), child_count), child_count):
            #missing required elements?
            child_node = self.get_child_node_by_idx(i)
            valid &= child_node.is_valid(None, errh)

        for syn in self.syntax:
            (bResult, err_str) = is_syntax_valid(seg_data, syn)
            if not bResult:
                syn_type = syn[0]
                if syn_type == 'E':
                    errh.ele_error('10', err_str, None, syn[1])
                else:
                    errh.ele_error('2', err_str, None, syn[1])
                valid &= False

        return valid

    def _split_syntax(self, syntax):
        """
        Split a Syntax string into a list
        """
        if syntax[0] not in ['P', 'R', 'C', 'L', 'E']:
            #self.logger.error('Syntax %s is not valid' % (syntax))
            return None
        syn = [syntax[0]]
        for i in range(len(syntax[1:]) / 2):
            syn.append(int(syntax[i * 2 + 1:i * 2 + 3]))
        return syn

    def get_cur_count(self):
        """
        @return: current count
        @rtype: int
        """
        raise DeprecationWarning('Moved to nodeCounter')
        return self._cur_count

    def incr_cur_count(self):
        raise DeprecationWarning('Moved to nodeCounter')
        self._cur_count += 1

    def reset_cur_count(self):
        """
        Set cur_count of node to zero
        """
        raise DeprecationWarning('Moved to nodeCounter')
        self._cur_count = 0

    def set_cur_count(self, ct):
        raise DeprecationWarning('Moved to nodeCounter')
        self._cur_count = ct

    def get_counts_list(self, ct_list):
        """
        Build a list of (path, ct) of the current node and parents
        Gets the node counts to apply to another map
        @param ct_list: List to append to
        @type ct_list: list[(string, int)]
        """
        raise DeprecationWarning('Moved to nodeCounter')
        my_ct = (self.get_path(), self._cur_count)
        ct_list.append(my_ct)
        if not self.parent.is_map_root():
            self.parent.get_counts_list(ct_list)
        return True

    def loop_segment_iterator(self):
        yield self


############################################################
# Element Interface
############################################################
class element_if(x12_node):
    """
    Element Interface
    """

    def __init__(self, root, parent, elem):
        """
        @param parent: parent node
        """
        x12_node.__init__(self)
        self.children = []
        self.root = root
        self.parent = parent
        self.base_name = 'element'
        self.valid_codes = []
        self.external_codes = None
        self.rec = None

        self.id = elem.get('xid')
        self.refdes = self.id
        self.data_ele = elem.get('data_ele') if elem.get(
            'data_ele') else elem.findtext('data_ele')
        self.usage = elem.get(
            'usage') if elem.get('usage') else elem.findtext('usage')
        self.name = elem.get(
            'name') if elem.get('name') else elem.findtext('name')
        self.seq = int(elem.get(
            'seq')) if elem.get('seq') else int(elem.findtext('seq'))
        self.path = elem.get(
            'seq') if elem.get('seq') else elem.findtext('seq')
        self.max_use = elem.get('max_use') if elem.get(
            'max_use') else elem.findtext('max_use')
        self.res = elem.findtext('regex')
        try:
            if self.res is not None and self.res != '':
                self.rec = re.compile(self.res, re.S)
        except Exception:
            raise EngineError('Element regex "%s" failed to compile' % (self.res))

        v = elem.find('valid_codes')
        if v is not None:
            self.external_codes = v.get('external')
            for c in v.findall('code'):
                self.valid_codes.append(c.text)

    def debug_print(self):
        sys.stdout.write(self.__repr__())
        for node in self.children:
            node.debug_print()

    def __repr__(self):
        """
        @rtype: string
        """
        data_ele = self.root.data_elements.get_by_elem_num(self.data_ele)
        out = '%s "%s"' % (self.refdes, self.name)
        if self.data_ele:
            out += '  data_ele: %s' % (self.data_ele)
        if self.usage:
            out += '  usage: %s' % (self.usage)
        if self.seq:
            out += '  seq: %i' % (self.seq)
        out += '  %s(%i, %i)' % (data_ele['data_type'], data_ele[
            'min_len'], data_ele['max_len'])
        if self.external_codes:
            out += '   external codes: %s' % (self.external_codes)
        out += '\n'
        return out

#    def __del__(self):
#        pass

    def _error(self, errh, err_str, err_cde, elem_val):
        """
        Forward the error to an error_handler
        """
        errh.ele_error(err_cde, err_str, elem_val, self.refdes)
            # pos=self.seq, data_ele=self.data_ele)

    def _valid_code(self, code):
        """
        Verify the x12 element value is in the given list of valid codes
        @return: True if found, else False
        @rtype: boolean
        """
        #if not self.valid_codes:
        #    return True
        if code in self.valid_codes:
            return True
        return False

    def get_parent(self):
        """
        @return: ref to parent class instance
        """
        return self.parent

    def is_match(self):
        """
        @return:
        @rtype: boolean
        """
        # match also by ID
        raise NotImplementedError('Override in sub-class')
        #return False

    def is_valid(self, elem, errh, type_list=[]):
        """
        Is this a valid element?

        @param elem: element instance
        @type elem: L{element<segment.Element>}
        @param errh: instance of error_handler
        @param check_dte: date string to check against (YYYYMMDD)
        @param type_list: Optional data/time type list
        @type type_list: list[string]
        @return: True if valid
        @rtype: boolean
        """
        errh.add_ele(self)

        if elem and elem.is_composite():
            err_str = 'Data element "%s" (%s) is an invalid composite' % \
                (self.name, self.refdes)
            self._error(errh, err_str, '6', elem.__repr__())
            return False
        if elem is None or elem.get_value() == '':
            if self.usage in ('N', 'S'):
                return True
            elif self.usage == 'R':
                if self.seq != 1 or not self.parent.is_composite() or self.parent.usage == 'R':
                    err_str = 'Mandatory data element "%s" (%s) is missing' % (
                        self.name, self.refdes)
                    self._error(errh, err_str, '1', None)
                    return False
                else:
                    return True
        if self.usage == 'N' and elem.get_value() != '':
            err_str = 'Data element "%s" (%s) is marked as Not Used' % (
                self.name, self.refdes)
            self._error(errh, err_str, '10', None)
            return False

        elem_val = elem.get_value()
        data_ele = self.root.data_elements.get_by_elem_num(self.data_ele)
        data_type = data_ele['data_type']
        min_len = data_ele['min_len']
        max_len = data_ele['max_len']
        valid = True
# Validate based on data_elem_num
# Then, validate on more specific criteria
        if (not data_type is None) and (data_type == 'R' or data_type[0] == 'N'):
            elem_strip = string.replace(
                string.replace(elem_val, '-', ''), '.', '')
            elem_len = len(elem_strip)
            if len(elem_strip) < min_len:
                err_str = 'Data element "%s" (%s) is too short: len("%s") = %i < %i (min_len)' % \
                    (self.name, self.refdes, elem_val, elem_len, min_len)
                self._error(errh, err_str, '4', elem_val)
                valid = False
            if len(elem_strip) > max_len:
                err_str = 'Data element "%s" (%s) is too long: len("%s") = %i > %i (max_len)' % \
                    (self.name, self.refdes, elem_val, elem_len, max_len)
                self._error(errh, err_str, '5', elem_val)
                valid = False
        else:
            elem_len = len(elem_val)
            if len(elem_val) < min_len:
                err_str = 'Data element "%s" (%s) is too short: len("%s") = %i < %i (min_len)' % \
                    (self.name, self.refdes, elem_val, elem_len, min_len)
                self._error(errh, err_str, '4', elem_val)
                valid = False
            if len(elem_val) > max_len:
                err_str = 'Data element "%s" (%s) is too long: len("%s") = %i > %i (max_len)' % \
                    (self.name, self.refdes, elem_val, elem_len, max_len)
                self._error(errh, err_str, '5', elem_val)
                valid = False

        (res, bad_string) = validation.contains_control_character(elem_val)
        if res:
            err_str = 'Data element "%s" (%s), contains an invalid control character(%s)' % \
                (self.name, self.refdes, bad_string)
            self._error(errh, err_str, '6', bad_string)
            valid = False
        if data_type in ['AN', 'ID'] and elem_val[-1] == ' ':
            if len(elem_val.rstrip()) >= min_len:
                err_str = 'Data element "%s" (%s) has unnecessary trailing spaces. (%s)' % \
                    (self.name, self.refdes, elem_val)
                self._error(errh, err_str, '6', elem_val)
                valid = False

        if not self._is_valid_code(elem_val, errh):
            valid = False
        if not validation.IsValidDataType(elem_val, data_type, self.root.param.get('charset'), self.root.icvn):
            if data_type in ('RD8', 'DT', 'D8', 'D6'):
                err_str = 'Data element "%s" (%s) contains an invalid date (%s)' % \
                    (self.name, self.refdes, elem_val)
                self._error(errh, err_str, '8', elem_val)
                valid = False
            elif data_type == 'TM':
                err_str = 'Data element "%s" (%s) contains an invalid time (%s)' % \
                    (self.name, self.refdes, elem_val)
                self._error(errh, err_str, '9', elem_val)
                valid = False
            else:
                err_str = 'Data element "%s" (%s) is type %s, contains an invalid character(%s)' % \
                    (self.name, self.refdes, data_type, elem_val)
                self._error(errh, err_str, '6', elem_val)
                valid = False
        if len(type_list) > 0:
            valid_type = False
            for dtype in type_list:
                valid_type |= validation.IsValidDataType(elem_val, dtype, self.root.param.get('charset'))
            if not valid_type:
                if 'TM' in type_list:
                    err_str = 'Data element "%s" (%s) contains an invalid time (%s)' % \
                        (self.name, self.refdes, elem_val)
                    self._error(errh, err_str, '9', elem_val)
                elif 'RD8' in type_list or 'DT' in type_list or 'D8' in type_list or 'D6' in type_list:
                    err_str = 'Data element "%s" (%s) contains an invalid date (%s)' % \
                        (self.name, self.refdes, elem_val)
                    self._error(errh, err_str, '8', elem_val)
                valid = False
        if self.rec:
            m = self.rec.search(elem_val)
            if not m:
                err_str = 'Data element "%s" with a value of (%s)' % \
                    (self.name, elem_val)
                err_str += ' failed to match the regular expression "%s"' % (
                    self.res)
                self._error(errh, err_str, '7', elem_val)
                valid = False
        return valid

    def _is_valid_code(self, elem_val, errh):
        """
        @rtype: boolean
        """
        bValidCode = False
        if len(self.valid_codes) == 0 and self.external_codes is None:
            bValidCode = True
        if elem_val in self.valid_codes:
            bValidCode = True
        if self.external_codes is not None and \
                self.root.ext_codes.isValid(self.external_codes, elem_val):
            bValidCode = True
        if not bValidCode:
            err_str = '(%s) is not a valid code for %s (%s)' % (
                elem_val, self.name, self.refdes)
            self._error(errh, err_str, '7', elem_val)
            return False
        return True

    def get_data_type(self):
        """
        """
        data_ele = self.root.data_elements.get_by_elem_num(self.data_ele)
        return data_ele['data_type']

    def get_seg_count(self):
        """
        """
        pass

    def is_element(self):
        """
        @rtype: boolean
        """
        return True


############################################################
# Composite Interface
############################################################
class composite_if(x12_node):
    """
    Composite Node Interface
    """
    def __init__(self, root, parent, elem):
        """
        Get the values for this composite
        @param parent: parent node
        """
        x12_node.__init__(self)

        self.children = []
        self.root = root
        self.parent = parent
        self.path = ''
        self.base_name = 'composite'

        self.id = elem.get('xid')
        self.refdes = elem.findtext(
            'refdes') if elem.findtext('refdes') else self.id
        self.data_ele = elem.get('data_ele') if elem.get(
            'data_ele') else elem.findtext('data_ele')
        self.usage = elem.get(
            'usage') if elem.get('usage') else elem.findtext('usage')
        self.seq = int(elem.get(
            'seq')) if elem.get('seq') else int(elem.findtext('seq'))
        self.repeat = int(elem.get('repeat')) if elem.get('repeat') else int(
            elem.findtext('repeat')) if elem.findtext('repeat') else 1
        self.name = elem.get(
            'name') if elem.get('name') else elem.findtext('name')

        for e in elem.findall('element'):
            self.children.append(element_if(self.root, self, e))

    def _error(self, errh, err_str, err_cde, elem_val):
        """
        Forward the error to an error_handler
        """
        err_str2 = err_str.replace('\n', '').replace('\r', '')
        elem_val2 = elem_val.replace('\n', '').replace('\r', '')
        errh.ele_error(err_cde, err_str2, elem_val2, self.refdes)
            #, pos=self.seq, data_ele=self.data_ele)

    def debug_print(self):
        sys.stdout.write(self.__repr__())
        for node in self.children:
            node.debug_print()

    def __repr__(self):
        """
        @rtype: string
        """
        out = '%s "%s"' % (self.id, self.name)
        if self.usage:
            out += '  usage %s' % (self.usage)
        if self.seq:
            out += '  seq %i' % (self.seq)
        if self.refdes:
            out += '  refdes %s' % (self.refdes)
        out += '\n'
        return out

    def xml(self):
        """
        Sends an xml representation of the composite to stdout
        """
        sys.stdout.write('<composite>\n')
        for sub_elem in self.children:
            sub_elem.xml()
        sys.stdout.write('</composite>\n')

    def is_valid(self, comp_data, errh):
        """
        Validates the composite
        @param comp_data: data composite instance, has multiple values
        @param errh: instance of error_handler
        @rtype: boolean
        """
        valid = True
        if (comp_data is None or comp_data.is_empty()) and self.usage in ('N', 'S'):
            return True

        if self.usage == 'R':
            good_flag = False
            for sub_ele in comp_data:
                if sub_ele is not None and len(sub_ele.get_value()) > 0:
                    good_flag = True
                    break
            if not good_flag:
                err_str = 'At least one component of composite "%s" (%s) is required' % \
                    (self.name, self.refdes)
                errh.ele_error('2', err_str, None, self.refdes)
                return False

        if self.usage == 'N' and not comp_data.is_empty():
            err_str = 'Composite "%s" (%s) is marked as Not Used' % (
                self.name, self.refdes)
            errh.ele_error('5', err_str, None, self.refdes)
            return False

        if len(comp_data) > self.get_child_count():
            err_str = 'Too many sub-elements in composite "%s" (%s)' % (
                self.name, self.refdes)
            errh.ele_error('3', err_str, None, self.refdes)
            valid = False
        for i in range(min(len(comp_data), self.get_child_count())):
            valid &= self.get_child_node_by_idx(i).is_valid(comp_data[i], errh)
        for i in range(min(len(comp_data), self.get_child_count()), self.get_child_count()):
            if i < self.get_child_count():
                #Check missing required elements
                valid &= self.get_child_node_by_idx(i).is_valid(None, errh)
        return valid

    def is_composite(self):
        """
        @rtype: boolean
        """
        return True


def load_map_file(map_file, param, map_path=None):
    """
    Create the map object from a file
    @param map_file: absolute path for file
    @type map_file: string
    @rtype: pyx12.map_if
    @param map_path: Override directory containing map xml files.  If None,
        uses package resource folder
    @type map_path: string
    """
    logger = logging.getLogger('pyx12')
    if map_path is not None:
        logger.debug("Looking for map file '{}' in map_path '{}'".format(map_file, map_path))
        map_fd = open(os.path.join(map_path, map_file))
    else:
        logger.debug("Looking for map file '{}' in pkg_resources".format(map_file))
        map_fd = resource_stream(__name__, os.path.join('map', map_file))
    imap = None
    try:
        logger.debug('Create map from %s' % (map_file))
        etree = et.parse(map_fd)
        imap = map_if(etree.getroot(), param, map_path)
    except AssertionError:
        logger.error('Load of map file failed: %s' % (map_file))
        raise
    except Exception:
        raise
        #raise EngineError('Load of map file failed: %s' % (map_file))
    return imap
